from unittest.mock import patch

from orchestra.communication.staffing import address_staffing_requests
from orchestra.models import Iteration
from orchestra.models import StaffBotRequest
from orchestra.models import StaffingRequestInquiry
from orchestra.models import Task
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.utils.task_lifecycle import assign_task
from orchestra.utils.task_lifecycle import create_subsequent_tasks
from orchestra.utils.task_lifecycle import submit_task


class StaffBotAutoAssignTestCase(OrchestraTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)

    @patch('orchestra.bots.staffbot.send_mail')
    @patch('orchestra.bots.staffbot.StaffBot._send_staffing_request_by_slack')
    def test_preassign_workers(self, mock_mail, mock_slack):
        request_cause = StaffBotRequest.RequestCause.TASK_POLICY.value
        staffing_request_count = StaffingRequestInquiry.objects.filter(
            request__request_cause=request_cause).count()
        project = self.projects['staffbot_assignment_policy']

        # Create first task in test project
        create_subsequent_tasks(project)
        address_staffing_requests()
        self.assertEqual(project.tasks.count(), 1)
        # Assign initial task to worker 0
        initial_task = assign_task(self.workers[0].id,
                                   project.tasks.first().id)
        # Submit task; next task should be created
        with patch('orchestra.utils.task_lifecycle._is_review_needed',
                   return_value=False):
            initial_task = submit_task(initial_task.id, {},
                                       Iteration.Status.REQUESTED_REVIEW,
                                       self.workers[0])

        # Mock mail should be called if we autostaff
        self.assertTrue(mock_mail.called)
        self.assertTrue(mock_slack.called)
        # Assert we created new StaffingRequestInquirys because of autostaff
        new_staffing_request_count = StaffingRequestInquiry.objects.filter(
            request__request_cause=request_cause).count()
        self.assertTrue(staffing_request_count < new_staffing_request_count)
        self.assertEqual(project.tasks.count(), 2)
        related_task = project.tasks.exclude(id=initial_task.id).first()
        # Worker 0 not certified for related tasks, so should not have been
        # auto-assigned
        self.assertEqual(related_task.assignments.count(), 0)
        self.assertEqual(related_task.status, Task.Status.AWAITING_PROCESSING)



from unittest.mock import MagicMock
from unittest.mock import patch

from django.conf import settings
from django.test import TestCase
from django.test import override_settings

from orchestra.google_apps.convenience import Service
from orchestra.google_apps.convenience import _get_image_mimetype
from orchestra.google_apps.convenience import add_image
from orchestra.google_apps.convenience import create_document_from_template
from orchestra.google_apps.convenience import create_folder_with_permissions
from orchestra.google_apps.convenience import create_media_folder_with_images
from orchestra.google_apps.errors import GoogleDriveError
from orchestra.google_apps.errors import InvalidUrlError
from orchestra.tests.helpers.google_apps import fake_image_get
from orchestra.tests.helpers.google_apps import mock_create_drive_service


@override_settings(GOOGLE_APPS=True)
class TestGoogleAppsConvenience(TestCase):

    def test_get_image_mimetype(self):
        # Content type is not provided.
        url = 'http://nocontenttype.com/image.jpeg'
        response = fake_image_get(url)
        mimetype = _get_image_mimetype(response, url)
        self.assertEqual(mimetype, 'image/jpeg')

        # Content type is provided.
        url = 'http://contenttype.com/image.jpg'
        response = fake_image_get(url)
        mimetype = _get_image_mimetype(response, url)
        self.assertEqual(mimetype, 'image/jpg')

    @patch('requests.get', side_effect=fake_image_get)
    @patch.object(Service, '_create_drive_service',
                  new=mock_create_drive_service)
    def test_add_image(self, requests_get):
        # Make sure if incorrect url then image is not returned.
        with self.assertRaises(InvalidUrlError):
            add_image(MagicMock(), 'test_folder', 'http://in.gogo/test.jp')

        service = Service(settings.GOOGLE_P12_PATH,
                          settings.GOOGLE_SERVICE_EMAIL)
        image_data = add_image(service, 'test_folder',
                               'http://nocontenttype.com/image.jpeg')
        self.assertEqual(
            image_data, {'id': 1, 'alternateLink': 'http://a.google.com/link'})

        image_data = add_image(service, 'test_folder',
                               'http://nocontenttype.com/error.jpg')
        self.assertIsNone(image_data)

    @patch('requests.get', side_effect=fake_image_get)
    @patch.object(Service, '_create_drive_service',
                  new=mock_create_drive_service)
    def test_create_media_folder(self, requests_get):
        with self.assertRaises(GoogleDriveError):
            create_media_folder_with_images('test_parent_id',
                                            ['http://in.gogo/image.jpg'],
                                            'error')

        # Make sure that if the correct image is provided,
        # it creates a folder.
        media_images = create_media_folder_with_images(
            'test_parent_id',
            ['http://inp.gogo/image.jpg'],
            'test')
        self.assertEqual(media_images['folder']['id'], 1)
        self.assertEqual(
            media_images['image_counter']['uploaded_images'], 1)

        # Make sure that if the incorrect image is provided,
        # still returns a new folder.
        media_images = create_media_folder_with_images(
            'test_parent_id',
            ['http://nocontenttype.com/test.jp', None],
            'test')
        self.assertEqual(media_images['folder']['id'], 1)
        self.assertEqual(
            media_images['image_counter']['not_uploaded_images'], 2)

        with self.assertRaises(GoogleDriveError):
            folder = create_folder_with_permissions(
                'test_parent_id', 'error')

        # In order to test permission failure we did something convoluted.
        # create_folder_with_permissions calls service.insert_folder which in
        # turn will create a new folder named 'error'.
        # When we try to add permission to folder 'error' that fails in
        # our helper function and raises an error.
        folder = create_folder_with_permissions('test_parent_id',
                                                'permission_fail')
        self.assertEqual(folder['id'], 'error')

    @patch.object(Service, '_create_drive_service',
                  new=mock_create_drive_service)
    def test_delete_folder(self):
        service = Service(settings.GOOGLE_P12_PATH,
                          settings.GOOGLE_SERVICE_EMAIL)
        folder = service.delete_folder('test')
        self.assertEqual(folder['id'], 1)

        folder = service.delete_folder('error')
        self.assertIsNone(folder)

    @patch.object(Service, '_create_drive_service',
                  new=mock_create_drive_service)
    def test_create_document_from_template(self):
        upload_info = create_document_from_template('test_id', 'test_filename')
        self.assertEqual(upload_info['id'], 1)
        with self.assertRaises(GoogleDriveError):
            create_document_from_template('error', 'test_filename')



class StaffingResponseException(Exception):
    pass


class SlackConnectionError(Exception):
    pass


class SlackUserUnauthorized(Exception):
    pass


class SlackCommandInvalidRequest(Exception):
    pass



import copy
import csv
import json

from dateutil.parser import parse
from django.urls import reverse
from io import StringIO
from unittest.mock import patch

from orchestra.models import Task
from orchestra.models import Todo
from orchestra.models import TodoQA
from orchestra.models import TodoListTemplate
from orchestra.models import TodoListTemplateImportRecord
from orchestra.models import Worker
from orchestra.project_api.serializers import TimeEntrySerializer
from orchestra.tests.helpers import EndpointTestCase
from orchestra.tests.helpers import OrchestraTransactionTestCase
from orchestra.tests.helpers.fixtures import TaskFactory
from orchestra.tests.helpers.fixtures import TodoFactory
from orchestra.tests.helpers.fixtures import StepFactory
from orchestra.tests.helpers.fixtures import ProjectFactory
from orchestra.tests.helpers.fixtures import TodoQAFactory
from orchestra.tests.helpers.fixtures import WorkflowVersionFactory
from orchestra.tests.helpers.fixtures import TodoListTemplateFactory
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.tests.helpers.fixtures import TODO_TEMPLATE_BAD_HEADER_CSV_TEXT
from orchestra.tests.helpers.fixtures import TODO_TEMPLATE_GOOD_CSV_TEXT
from orchestra.tests.helpers.fixtures import (
    TODO_TEMPLATE_INVALID_PARENT_CSV_TEXT)
from orchestra.tests.helpers.fixtures import TODO_TEMPLATE_NESTED_TODOS
from orchestra.tests.helpers.fixtures import TODO_TEMPLATE_TWO_ENTRIES_CSV_TEXT
from orchestra.todos.serializers import TodoQASerializer
from orchestra.todos.serializers import BulkTodoSerializerWithoutQA
from orchestra.todos.serializers import TodoListTemplateSerializer
from orchestra.utils.load_json import load_encoded_json


def _todo_data(title, status=Todo.Status.PENDING.value,
               start_by=None,
               due=None, parent_todo=None, template=None,
               activity_log=str({'actions': []}), qa=None,
               project=None, step=None, details=None, is_deleted=False,
               slug=None, required=False):
    return {
        'title': title,
        'template': template,
        'parent_todo': parent_todo,
        'start_by_datetime': start_by,
        'due_datetime': due,
        'activity_log': activity_log,
        'qa': qa,
        'additional_data': {},
        'order': None,
        'project': project,
        'section': None,
        'status': status,
        'step': step,
        'slug': slug,
        'details': details,
        'is_deleted': is_deleted,
        'required': required
    }


def _get_test_conditional_props(project):
    return {
        'prop1': True,
        'prop2': False
    }


class TodosEndpointTests(EndpointTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.worker = Worker.objects.get(user__username='test_user_6')
        self.request_client.login(username=self.worker.user.username,
                                  password='defaultpassword')
        self.list_create_url = reverse('orchestra:todos:todos-list')
        self.list_details_url_name = 'orchestra:todos:todos-detail'
        self.workflow_version = WorkflowVersionFactory()
        self.step = StepFactory(
            slug='step-slug',
            workflow_version=self.workflow_version)
        self.project = ProjectFactory(
            workflow_version=self.workflow_version)
        tasks = Task.objects.filter(assignments__worker=self.worker)
        task = tasks[0]
        task.project = self.project
        task.step = self.step
        task.save()
        self.todo_title = 'Let us do this'
        self.deadline = parse('2018-01-16T07:03:00+00:00')

    def _verify_missing_task(self, response):
        self.assertEqual(response.status_code, 400)
        data = load_encoded_json(response.content)
        self.assertEqual(data['message'], 'No task for given id')
        self.assertEqual(data['error'], 400)

    def _verify_worker_not_assigned(self, response):
        self.assertEqual(response.status_code, 400)
        data = load_encoded_json(response.content)
        self.assertEqual(data['message'],
                         'Worker is not assigned to this task id.')
        self.assertEqual(data['error'], 400)

    def _verify_time_entries(self, data):
        for time_entry in data:
            serializer = TimeEntrySerializer(data=time_entry)
            self.assertTrue(serializer.is_valid())

    def _verify_todo_content(self, todo, expected_todo):
        todo = dict(todo)
        created_at = todo.pop('created_at')
        todo_id = todo.pop('id')
        self.assertEqual(todo, expected_todo)
        self.assertGreater(len(created_at), 0)
        self.assertGreaterEqual(todo_id, 0)

    def _verify_todos_list(self, project_id, expected_todos, success):
        resp = self.request_client.get(self.list_create_url,
                                       {'project__id': project_id})
        if success:
            self.assertEqual(resp.status_code, 200)
            data = load_encoded_json(resp.content)
            for todo, expected_todo in zip(data, expected_todos):
                self._verify_todo_content(todo, expected_todo)
        else:
            self.assertEqual(resp.status_code, 403)

    @patch('orchestra.todos.views.notify_todo_created')
    def _verify_todo_creation(self, success, project, step, mock_notify):
        num_todos = Todo.objects.all().count()
        resp = self.request_client.post(self.list_create_url, {
            'project': project,
            'step': step.slug,
            'title': self.todo_title,
            'status': Todo.Status.PENDING.value})
        if success:
            self.assertEqual(resp.status_code, 201)
            self.assertEqual(Todo.objects.all().count(), num_todos + 1)
            todo = load_encoded_json(resp.content)
            self._verify_todo_content(
                todo, _todo_data(
                    self.todo_title, project=project, step=step.slug))
            self.assertTrue(mock_notify.called)
        else:
            self.assertEqual(resp.status_code, 403)
            self.assertEqual(Todo.objects.all().count(), num_todos)

    @patch('orchestra.todos.views.notify_single_todo_update')
    def _verify_todo_update(self, todo, success, mock_notify):
        title = 'new title'
        list_details_url = reverse(
            self.list_details_url_name,
            kwargs={'pk': todo.id})
        resp = self.request_client.put(
            list_details_url,
            json.dumps(_todo_data(
                title, status=Todo.Status.COMPLETED.value,
                project=self.project.id, step=self.step.slug)),
            content_type='application/json')
        updated_todo = BulkTodoSerializerWithoutQA(
            Todo.objects.get(id=todo.id)).data
        if success:
            self.assertEqual(resp.status_code, 200)
            self._verify_todo_content(
                updated_todo, _todo_data(
                    title, status=Todo.Status.COMPLETED.value,
                    project=self.project.id,
                    step=self.step.slug))
            self.assertTrue(mock_notify.called)
        else:
            self.assertEqual(resp.status_code, 403)
            self.assertNotEqual(updated_todo['title'], title)

    def test_todos_list_create(self):
        self._verify_todos_list(self.project.id, [], True)
        self._verify_todo_creation(
            True, self.project.id, self.step)
        self._verify_todos_list(self.project.id,
                                [_todo_data(
                                    self.todo_title,
                                    project=self.project.id,
                                    step=self.step.slug)],
                                True)

    def test_todos_list_create_permissions(self):
        # Can't make requests for projects in which you're uninvolved.
        project = ProjectFactory()
        step = StepFactory()
        self._verify_todos_list(project.id, [], False)
        self._verify_todo_creation(False, project.id, step)

    def test_todo_details_and_permissions(self):
        # You should be able to update Todos for projects in which
        # you're involved, and not for other projects.
        good_todo = TodoFactory(project=self.project, step=self.step)
        self._verify_todo_update(good_todo, True)
        bad_todo = TodoFactory()
        self._verify_todo_update(bad_todo, False)

    def test_create_todo_with_start_by_datetime(self):
        START_TITLE = 'Start soon'

        start_by_todo = TodoFactory(
            project=self.project,
            step=self.step,
            start_by_datetime=self.deadline,
            title=START_TITLE,
            status=Todo.Status.PENDING.value)

        self._verify_todos_list(start_by_todo.project.id, [
            _todo_data(
                START_TITLE,
                status=Todo.Status.PENDING.value,
                start_by=self.deadline.strftime('%Y-%m-%dT%H:%M:%SZ'),
                due=None,
                project=start_by_todo.project.id,
                step=start_by_todo.step.slug)
        ], True)

    def test_create_todo_with_due_datetime(self):
        DUE_TITLE = 'Due soon'

        due_todo = TodoFactory(
            project=self.project,
            step=self.step,
            due_datetime=self.deadline,
            title=DUE_TITLE,
            status=Todo.Status.PENDING.value)

        self._verify_todos_list(due_todo.project.id, [
            _todo_data(
                DUE_TITLE,
                start_by=None,
                due=self.deadline.strftime('%Y-%m-%dT%H:%M:%SZ'),
                project=due_todo.project.id,
                step=due_todo.step.slug),
        ], True)

    def test_todo_deletion(self):
        todo = TodoFactory(project=self.project, step=self.step)
        deletion_url = reverse(
            self.list_details_url_name,
            kwargs={'pk': todo.id})
        resp = self.request_client.delete(deletion_url)
        self.assertEqual(resp.status_code, 204)


class TodoQAEndpointTests(EndpointTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.worker = Worker.objects.get(user__username='test_user_6')
        self.request_client.login(username=self.worker.user.username,
                                  password='defaultpassword')
        self.list_create_url = reverse('orchestra:todos:todo_qas')
        self.worker_task_recent_todo_qas_url = reverse(
            'orchestra:todos:worker_task_recent_todo_qas')
        self.list_details_url_name = 'orchestra:todos:todo_qa'
        self.project0 = ProjectFactory()
        self.project1 = ProjectFactory()
        self.step = StepFactory(slug='some-slug')
        tasks = Task.objects.filter(
            assignments__worker=self.worker)
        self.task_0 = tasks[0]
        self.task_0.project = self.project0
        self.task_0.step = self.step
        self.task_0.save()
        self.task_1 = tasks[1]
        self.task_1.project = self.project1
        self.task_1.step = self.step
        self.task_1.save()
        self.todo = TodoFactory(project=self.project0, step=self.step)
        self.comment = 'Test comment'

    def _todo_qa_data(
            self, todo, approved, comment):
        return {
            'todo': todo.id,
            'approved': approved,
            'comment': comment
        }

    def _verify_todo_qa_content(self, todo_qa,
                                expected_todo_qa):
        todo_qa = dict(todo_qa)
        created_at = todo_qa.pop('created_at')
        todo_qa_id = todo_qa.pop('id')
        self.assertEqual(todo_qa, expected_todo_qa)
        self.assertGreater(len(created_at), 0)
        self.assertGreaterEqual(todo_qa_id, 0)

    def _verify_todo_qa_creation(self, todo, success):
        num_todo_qas = TodoQA.objects.all().count()
        resp = self.request_client.post(self.list_create_url, {
            'todo': todo.id,
            'approved': True,
            'comment': self.comment})
        if success:
            self.assertEqual(resp.status_code, 201)
            self.assertEqual(TodoQA.objects.all().count(), num_todo_qas + 1)
            todo_qa = load_encoded_json(resp.content)
            self._verify_todo_qa_content(todo_qa, self._todo_qa_data(
                todo, True, self.comment))
        else:
            self.assertEqual(resp.status_code, 403)
            self.assertEqual(TodoQA.objects.all().count(), num_todo_qas)

    def _verify_todo_qa_update(self, todo_qa, success):
        comment = 'new comment'
        list_details_url = reverse(
            self.list_details_url_name,
            kwargs={'pk': todo_qa.id})
        resp = self.request_client.put(
            list_details_url,
            json.dumps(self._todo_qa_data(
                todo_qa.todo, True, comment)),
            content_type='application/json')
        updated_todo_qa = TodoQASerializer(
            TodoQA.objects.get(id=todo_qa.id)).data
        if success:
            self.assertEqual(resp.status_code, 200)
            self._verify_todo_qa_content(updated_todo_qa, self._todo_qa_data(
                todo_qa.todo, True, comment))
        else:
            self.assertEqual(resp.status_code, 403)
            self.assertNotEqual(
                updated_todo_qa['comment'], comment)

    def test_todo_qas_list_create(self):
        self._verify_todo_qa_creation(self.todo, True)

    def test_todo_qas_create_permissions(self):
        # Can't make requests for projects in which you're uninvolved.
        project = ProjectFactory()
        step = StepFactory()
        todo = TodoFactory(project=project, step=step)
        self._verify_todo_qa_creation(todo, False)

    def test_todo_qa_details_and_permissions(self):
        # You should be able to update TodoQAs for projects in which
        # you're involved, and not for other projects.
        good_todo_qa = TodoQAFactory(todo=self.todo)
        self._verify_todo_qa_update(good_todo_qa, True)
        bad_todo_qa = TodoQAFactory()
        self._verify_todo_qa_update(bad_todo_qa, False)

    def _verify_worker_task_recent_todo_qas(self, task, todo_qa, success):
        resp = self.request_client.get(self.worker_task_recent_todo_qas_url,
                                       {'task': task.id})
        if success:
            self.assertEqual(resp.status_code, 200)
            data = load_encoded_json(resp.content)
            if todo_qa:
                self.assertEqual(
                    TodoQASerializer(todo_qa).data,
                    list(data.values())[0])
                self.assertEqual(1, len(data.keys()))
            else:
                self.assertEqual({}, data)
        else:
            self.assertEqual(resp.status_code, 403)

    def test_worker_task_recent_todo_qas(self):
        todo_task_0 = TodoFactory(
            project=self.project0, step=self.step)
        todo_task_1 = TodoFactory(
            project=self.project1, step=self.step)

        # Zero TodoQAs
        self._verify_worker_task_recent_todo_qas(
            self.task_0, None, True)

        todo_qa_task_0 = TodoQAFactory(todo=todo_task_0, approved=False)

        # Most recent TodoQA is todo_qa_task_0
        self._verify_worker_task_recent_todo_qas(
            self.task_0, todo_qa_task_0, True)

        self._verify_worker_task_recent_todo_qas(
            self.task_1, todo_qa_task_0, True)

        todo_qa_task_1 = TodoQAFactory(todo=todo_task_1, approved=False)

        # If available use the todo qa for the corresponding task.
        self._verify_worker_task_recent_todo_qas(
            self.task_0, todo_qa_task_0, True)

        self._verify_worker_task_recent_todo_qas(
            self.task_1, todo_qa_task_1, True)

        todo_qa_task_0.delete()

        # Most recent TodoQA is todo_qa_task_1
        self._verify_worker_task_recent_todo_qas(
            self.task_0, todo_qa_task_1, True)

        self._verify_worker_task_recent_todo_qas(
            self.task_1, todo_qa_task_1, True)

        # Can't make requests for projects in which you're uninvolved.
        bad_task = TaskFactory()
        todo_bad_task = TodoFactory(task=bad_task)
        todo_qa_bad_task = TodoQAFactory(todo=todo_bad_task, approved=False)
        self._verify_worker_task_recent_todo_qas(
            bad_task, todo_qa_bad_task, False)


class TodoTemplateEndpointTests(EndpointTestCase):
    def setUp(self):
        super().setUp()
        setup_models(self)
        self.worker = Worker.objects.get(user__username='test_user_6')
        self.request_client.login(username=self.worker.user.username,
                                  password='defaultpassword')
        self.todolist_template_list_url = reverse(
            'orchestra:todos:todolist_templates')
        self.todolist_template_detail_url_name = \
            'orchestra:todos:todolist_template'
        self.todolist_template_slug = 'test_todolist_template_slug'
        self.todolist_template_name = 'test_todolist_template_name'
        self.todolist_template_description = \
            'test_todolist_template_description'
        self.workflow_version = WorkflowVersionFactory()
        self.step = StepFactory(
            slug='step-slug',
            workflow_version=self.workflow_version)
        self.slug = 'todo-item-slug'
        self.project = ProjectFactory(
            workflow_version=self.workflow_version)
        self.project2 = ProjectFactory()
        tasks = Task.objects.filter(assignments__worker=self.worker)
        self.task = tasks[0]
        self.task.project = self.project
        self.task.step = self.step
        self.task.save()

    def _todolist_template_data(
            self, slug, name, description, creator=None,
            todos="{'items': []}"):
        return {
            'slug': slug,
            'name': name,
            'description': description,
            'creator': creator,
            'todos': todos
        }

    def _verify_todolist_template_content(self, todolist_template,
                                          expected_todolist_template):
        todolist_template = dict(todolist_template)
        created_at = todolist_template.pop('created_at')
        todolist_template_id = todolist_template.pop('id')
        self.assertEqual(todolist_template, expected_todolist_template)
        self.assertGreater(len(created_at), 0)
        self.assertGreaterEqual(todolist_template_id, 0)

    def _verify_todolist_template_list(self, expected_todolist_templates):
        resp = self.request_client.get(self.todolist_template_list_url)
        self.assertEqual(resp.status_code, 200)
        length = len(expected_todolist_templates)
        data = load_encoded_json(resp.content)
        for todolist_template, expected_todolist_template in \
                zip(data[-length:], expected_todolist_templates):
            self._verify_todolist_template_content(
                todolist_template, expected_todolist_template)

    def _verify_todolist_template_creation(self):
        num_todolist_templates = TodoListTemplate.objects.all().count()
        resp = self.request_client.post(self.todolist_template_list_url, {
            'slug': self.todolist_template_slug,
            'name': self.todolist_template_name,
            'description': self.todolist_template_description})

        self.assertEqual(resp.status_code, 201)
        self.assertEqual(TodoListTemplate.objects.all().count(),
                         num_todolist_templates + 1)
        todolist_template = load_encoded_json(resp.content)
        self._verify_todolist_template_content(
            todolist_template, self._todolist_template_data(
                self.todolist_template_slug,
                self.todolist_template_name,
                self.todolist_template_description))

    def test_todolist_template_update(self):
        todolist_template = TodoListTemplateFactory()
        updated_description = 'updated description'
        todolist_template_detail_url = reverse(
            self.todolist_template_detail_url_name,
            kwargs={'pk': todolist_template.id})
        resp = self.request_client.put(
            todolist_template_detail_url,
            json.dumps(self._todolist_template_data(
                todolist_template.slug,
                todolist_template.name,
                updated_description)),
            content_type='application/json')
        updated_todolist_template = TodoListTemplateSerializer(
            TodoListTemplate.objects.get(id=todolist_template.id)).data

        self.assertEqual(resp.status_code, 200)
        self._verify_todolist_template_content(
            updated_todolist_template, self._todolist_template_data(
                todolist_template.slug,
                todolist_template.name,
                updated_description))

    def test_todolist_template_list_create(self):
        self._verify_todolist_template_list([])
        self._verify_todolist_template_creation()
        self._verify_todolist_template_list(
            [self._todolist_template_data(
                self.todolist_template_slug,
                self.todolist_template_name,
                self.todolist_template_description)])

    def _verify_todo_content(self, todo, expected_todo):
        todo = dict(todo)
        created_at = todo.pop('created_at')
        todo_id = todo.pop('id')

        self.assertEqual(todo, expected_todo)
        self.assertGreater(len(created_at), 0)
        self.assertGreaterEqual(todo_id, 0)

    def test_update_todos_from_todolist_template_success(self):
        num_todos = Todo.objects.all().count()
        update_todos_from_todolist_template_url = \
            reverse('orchestra:todos:update_todos_from_todolist_template')
        todolist_template = TodoListTemplateFactory(
            slug=self.todolist_template_slug,
            name=self.todolist_template_name,
            description=self.todolist_template_description,
            todos={'items': [{
                'id': 1,
                'description': 'todo parent',
                'project': self.project.id,
                'step': self.step.slug,
                'slug': None,
                'items': [{
                    'id': 2,
                    'project': self.project.id,
                    'step': self.step.slug,
                    'slug': self.slug,
                    'description': 'todo child',
                    'items': []
                }]
            }]},
        )
        resp = self.request_client.post(
            update_todos_from_todolist_template_url,
            {
                'todolist_template': todolist_template.slug,
                'project': self.project.id,
                'step': self.step.slug
            })

        self.assertEqual(resp.status_code, 200)
        self.assertEqual(Todo.objects.all().count(), num_todos + 3)
        todos = load_encoded_json(resp.content)
        expected_todos = [
            _todo_data('todo child',
                       template=todolist_template.id,
                       parent_todo=todos[1]['id'],
                       project=self.project.id,
                       step=self.step.slug,
                       slug=self.slug),
            _todo_data('todo parent',
                       template=todolist_template.id,
                       parent_todo=todos[2]['id'],
                       project=self.project.id,
                       step=self.step.slug),
            _todo_data(self.todolist_template_name,
                       template=todolist_template.id,
                       project=self.project.id,
                       step=self.step.slug),
        ]
        for todo, expected_todo in zip(todos, expected_todos):
            self._verify_todo_content(todo, expected_todo)

    def test_update_todos_from_todolist_template_missing_project_id(self):
        update_todos_from_todolist_template_url = \
            reverse('orchestra:todos:update_todos_from_todolist_template')
        todolist_template = TodoListTemplateFactory(
            slug=self.todolist_template_slug,
            name=self.todolist_template_name,
            description=self.todolist_template_description,
            todos={'items': []},
        )
        resp = self.request_client.post(
            update_todos_from_todolist_template_url,
            {
                'todolist_template': todolist_template.slug,
                'step': self.step.slug
            })

        self.assertEqual(resp.status_code, 403)

    def test_update_todos_from_todolist_template_invalid_project(self):
        update_todos_from_todolist_template_url = \
            reverse('orchestra:todos:update_todos_from_todolist_template')
        todolist_template = TodoListTemplateFactory(
            slug=self.todolist_template_slug,
            name=self.todolist_template_name,
            description=self.todolist_template_description,
            todos={'items': []},
        )
        resp = self.request_client.post(
            update_todos_from_todolist_template_url,
            {
                'todolist_template': todolist_template.slug,
                'project': self.project2.id,
                'step': self.step.id
            })

        self.assertEqual(resp.status_code, 403)

    def test_update_todos_from_todolist_template_invalid_todolist_template(
            self):
        update_todos_from_todolist_template_url = \
            reverse('orchestra:todos:update_todos_from_todolist_template')
        resp = self.request_client.post(
            update_todos_from_todolist_template_url,
            {
                'todolist_template': 'invalid-slug',
                'project': self.project.id,
                'step': self.step.id
            })

        self.assertEqual(resp.status_code, 400)

    def test_conditional_skip_remove_todos_from_template(self):
        update_todos_from_todolist_template_url = \
            reverse('orchestra:todos:update_todos_from_todolist_template')

        todolist_template = TodoListTemplateFactory(
            slug=self.todolist_template_slug,
            name=self.todolist_template_name,
            description=self.todolist_template_description,
            conditional_property_function={
                'path': 'orchestra.tests.test_todos'
                        '._get_test_conditional_props'
            },
            todos={'items': [
                {
                    'id': 1,
                    'description': 'todo parent 1',
                    'project': self.project.id,
                    'slug': None,
                    'items': [{
                        'id': 2,
                        'description': 'todo child 1',
                        'slug': None,
                        'project': self.project.id,
                        'items': []
                    }],
                    'remove_if': [{
                        'prop1': {
                            'operator': '==',
                            'value': True
                        }
                    }]
                }, {
                    'id': 3,
                    'description': 'todo parent 2',
                    'project': self.project.id,
                    'slug': None,
                    'items': [{
                        'id': 4,
                        'description': 'todo child 2',
                        'slug': None,
                        'project': self.project.id,
                        'items': [],
                        'skip_if': [{
                            'prop2': {
                                'operator': '!=',
                                'value': True
                            }
                        }]
                    }]
                }]},
        )
        resp = self.request_client.post(
            update_todos_from_todolist_template_url,
            {
                'todolist_template': todolist_template.slug,
                'project': self.project.id,
                'step': self.step.slug
            })
        self.assertEqual(resp.status_code, 200)
        todos = load_encoded_json(resp.content)

        expected_todos = [
            _todo_data('todo child 2',
                       template=todolist_template.id,
                       parent_todo=todos[1]['id'],
                       status=Todo.Status.DECLINED.value,
                       project=self.project.id,
                       step=self.step.slug),
            _todo_data('todo parent 2',
                       template=todolist_template.id,
                       parent_todo=todos[2]['id'],
                       project=self.project.id,
                       step=self.step.slug),
            _todo_data(self.todolist_template_name,
                       template=todolist_template.id,
                       project=self.project.id,
                       step=self.step.slug),
        ]
        for todo, expected_todo in zip(todos, expected_todos):
            self._verify_todo_content(todo, expected_todo)


class TodoListTemplatesImportExportTests(OrchestraTransactionTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.worker = Worker.objects.get(user__username='test_user_6')
        self.worker.user.is_staff = True
        self.worker.user.save()
        self.request_client.login(username=self.worker.user.username,
                                  password='defaultpassword')
        self.full_todo_list_template = TodoListTemplateFactory(
            slug='full-template-slug',
            name='Full template name',
            description='Full template description',
            conditional_property_function={
                'path': 'orchestra.tests.test_todos'
                        '._get_test_conditional_props'
            },
            todos=TODO_TEMPLATE_NESTED_TODOS
        )
        self.empty_todo_list_template = TodoListTemplateFactory(
            slug='empty-template-slug',
            name='Empty template name',
            description='Empty template description',
            conditional_property_function={
                'path': 'orchestra.tests.test_todos'
                        '._get_test_conditional_props'
            },
            todos={}
        )

        self.import_url_name = (
            'orchestra:todos:import_todo_list_template_from_spreadsheet')
        self.template_admin_name = 'admin:orchestra_todolisttemplate_change'
        self.export_url_name = 'admin:orchestra_todolisttemplate_actions'

    @patch('orchestra.todos.import_export._upload_csv_to_google')
    def test_export_spreadsheet(self, mock_upload):
        mock_upload.return_value = 'https://redirect.com/the_spreadsheet'
        export_url = reverse(
            self.export_url_name,
            kwargs={'pk': self.full_todo_list_template.id,
                    'tool': 'export_spreadsheet'})
        response = self.request_client.get(export_url)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, 'https://redirect.com/the_spreadsheet')
        self.assertEqual(mock_upload.call_count, 1)
        title_prefix = (mock_upload.call_args[0][0]
                        [:len(self.full_todo_list_template.name) + 2])
        self.assertEqual(
            title_prefix,
            '{} -'.format(self.full_todo_list_template.name))
        with open(mock_upload.call_args[0][1].name, 'r') as exported_file:
            lines = ''.join(exported_file.readlines())
            # The JSON encoding of properties is nondeterministic in order.
            lines = lines.replace(
                '"[{""prop"": {""operator"": ""=="", ""value"": true}}]"',
                'PROPREPLACED')
            lines = lines.replace(
                '"[{""prop"": {""value"": true, ""operator"": ""==""}}]"',
                'PROPREPLACED')
            correct_text = TODO_TEMPLATE_GOOD_CSV_TEXT.replace(
                '"[{""prop"": {""value"": true, ""operator"": ""==""}}]"',
                'PROPREPLACED')
            self.assertEqual(lines, correct_text)

    @patch('orchestra.todos.import_export.get_google_spreadsheet_as_csv')
    def test_import_spreadsheet(self, mock_get_spreadsheet):
        mock_get_spreadsheet.side_effect = self._fake_get_spreadsheet(
            TODO_TEMPLATE_GOOD_CSV_TEXT)

        import_url = reverse(
            self.import_url_name,
            kwargs={'pk': self.empty_todo_list_template.id})
        admin_url = reverse(
            self.template_admin_name,
            kwargs={'object_id': self.empty_todo_list_template.id})

        response = self.request_client.get(import_url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(
            response,
            'orchestra/import_todo_list_template_from_spreadsheet.html')
        self.assertEqual(mock_get_spreadsheet.call_count, 0)

        # Before import, the empty_todo_list_template should have no
        # template to-dos. After import, it should have the same
        # to-dos as full_todo_list_template (the CSV we load from is
        # equivalent to full_todo_list_template's to-dos).
        self.empty_todo_list_template.refresh_from_db()
        self.assertEqual(self.empty_todo_list_template.todos, {})
        self.assertEqual(
            TodoListTemplateImportRecord.objects.all().count(),
            0)
        response = self.request_client.post(
            import_url,
            {'spreadsheet_url': 'https://the-spreadsheet.com/url'})
        self.assertRedirects(
            response, admin_url,
            target_status_code=403)  # User doesn't have model access.
        self.assertEqual(mock_get_spreadsheet.call_count, 1)
        self.assertEqual(mock_get_spreadsheet.call_args[0][0],
                         'https://the-spreadsheet.com/url')
        self.empty_todo_list_template.refresh_from_db()

        # Walk the to-do tree, moving IDs (which won't be equal, but
        # should be unique) into a set. Ensure the rest of the fields
        # are equivalent after setting missing `skip_id`/`remove_if`
        # to `[]`.
        def recursively_pop_ids(todos, id_set):
            id = todos.pop('id', None)
            if id is not None:
                id_set.add(id)
            todos.setdefault('remove_if', [])
            todos.setdefault('skip_if', [])
            for item in todos.get('items', []):
                recursively_pop_ids(item, id_set)
        empty_ids = set()
        full_ids = set()
        empty_todos = copy.deepcopy(self.empty_todo_list_template.todos)
        full_todos = copy.deepcopy(self.full_todo_list_template.todos)
        recursively_pop_ids(empty_todos, empty_ids)
        recursively_pop_ids(full_todos, full_ids)
        self.assertEqual(len(empty_ids), 7)
        self.assertEqual(len(empty_ids), len(full_ids))
        self.assertEqual(empty_todos, full_todos)

        self.assertEqual(
            TodoListTemplateImportRecord.objects.all().count(),
            1)
        import_record = TodoListTemplateImportRecord.objects.first()
        self.assertEqual(import_record.todo_list_template.id,
                         self.empty_todo_list_template.id)
        self.assertEqual(import_record.importer.id,
                         self.worker.user.id)
        self.assertEqual(import_record.import_url,
                         'https://the-spreadsheet.com/url')

    @patch('orchestra.todos.import_export.get_google_spreadsheet_as_csv')
    def test_import_spreadsheet_errors(self, mock_get_spreadsheet):
        import_url = reverse(
            self.import_url_name,
            kwargs={'pk': self.empty_todo_list_template.id})

        mock_get_spreadsheet.side_effect = self._fake_get_spreadsheet(
            TODO_TEMPLATE_BAD_HEADER_CSV_TEXT)
        response = self.request_client.post(
            import_url,
            {'spreadsheet_url': 'https://the-spreadsheet.com/url'})
        self.assertEqual(response.status_code, 200)  # We didn't redirect.
        self.assertEqual(mock_get_spreadsheet.call_count, 1)
        self.assertRegex(response.content.decode('utf-8'),
                         'Error: Unexpected header:')
        mock_get_spreadsheet.reset_mock()

        mock_get_spreadsheet.side_effect = self._fake_get_spreadsheet(
            TODO_TEMPLATE_TWO_ENTRIES_CSV_TEXT)
        response = self.request_client.post(
            import_url,
            {'spreadsheet_url': 'https://the-spreadsheet.com/url'})
        self.assertEqual(response.status_code, 200)  # We didn't redirect.
        self.assertEqual(mock_get_spreadsheet.call_count, 1)
        self.assertRegex(response.content.decode('utf-8'),
                         'Error: More than one text entry in row 0: ')
        mock_get_spreadsheet.reset_mock()

        mock_get_spreadsheet.side_effect = self._fake_get_spreadsheet(
            TODO_TEMPLATE_INVALID_PARENT_CSV_TEXT)
        response = self.request_client.post(
            import_url,
            {'spreadsheet_url': 'https://the-spreadsheet.com/url'})
        self.assertEqual(response.status_code, 200)  # We didn't redirect.
        self.assertEqual(mock_get_spreadsheet.call_count, 1)
        self.assertRegex(response.content.decode('utf-8'),
                         'Error: Row 1 has skipped some columns in depth: ')
        mock_get_spreadsheet.reset_mock()

        # Nothing should have been created/updated since every import
        # failed.
        self.empty_todo_list_template.refresh_from_db()
        self.assertEqual(self.empty_todo_list_template.todos, {})
        self.assertEqual(
            TodoListTemplateImportRecord.objects.all().count(),
            0)

    def _fake_get_spreadsheet(self, csv_text):
        def fake_get_spreadsheet(spreadsheet_url, reader):
            return reader(
                StringIO(csv_text), dialect=csv.excel)
        return fake_get_spreadsheet



# Generated by Django 2.2.13 on 2020-09-21 17:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0089_todolisttemplateimportrecord'),
    ]

    operations = [
        migrations.AddField(
            model_name='step',
            name='completion_ends_project',
            field=models.BooleanField(default=False),
        ),
    ]



import json
import os
from copy import deepcopy

from django.db import transaction
from django.test import modify_settings

from orchestra.core.errors import WorkflowError
from orchestra.models import Workflow
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.workflow import assert_test_dir_v1_loaded
from orchestra.tests.helpers.workflow import assert_test_dir_v1_not_loaded
from orchestra.tests.helpers.workflow import assert_test_dir_v2_loaded
from orchestra.tests.helpers.workflow import assert_test_dir_v2_not_loaded
from orchestra.tests.helpers.workflow import assert_test_dir_workflow_loaded
from orchestra.tests.helpers.workflow import \
    assert_test_dir_workflow_not_loaded  # noqa
from orchestra.workflow.load import load_workflow
from orchestra.workflow.load import load_workflow_version

TEST_WORKFLOWS_MODULE = 'orchestra.tests.workflows'
TEST_WORKFLOWS_DIR = os.path.abspath(
    os.path.join(os.path.dirname(__file__), '../workflows'))
WORKFLOWS = {
    'bad_certifications': {
        'dir': os.path.join(TEST_WORKFLOWS_DIR, 'bad_certification_dir'),
        'app_label': 'bad_certification_dir',
    },
    'valid': {
        'dir': os.path.join(TEST_WORKFLOWS_DIR, 'test_dir'),
        'app_label': 'test_dir',
    }
}
WORKFLOW_MODULES = [
    '{}.{}'.format(TEST_WORKFLOWS_MODULE, workflow_dict['app_label'])
    for workflow_dict in WORKFLOWS.values()
]
INVALID_VERSION = 'INVALID'
VERSION_1 = 'test_v1'
VERSION_2 = 'test_v2'


@modify_settings(INSTALLED_APPS={
    'append': WORKFLOW_MODULES,
})
class LoadWorkflowTestCase(OrchestraTestCase):

    def test_load_workflow(self):
        """ Ensure that a workflow directory loads correctly. """
        # Verify initial DB state.
        assert_test_dir_workflow_not_loaded(self)
        assert_test_dir_v1_not_loaded(self)
        assert_test_dir_v2_not_loaded(self)

        # Shouldn't be able to parse a workflow with nonexistent
        # certification dependencies.
        with self.assertRaisesMessage(
                WorkflowError,
                'Certification certification1 requires non-existent '
                'certification'):
            load_workflow(WORKFLOWS['bad_certifications']['app_label'],
                          VERSION_1)

        # Shouldn't be able to load a non-existent version.
        version_error = 'Invalid version requested: {}'.format(
            INVALID_VERSION)
        with self.assertRaisesMessage(WorkflowError, version_error):
            load_workflow(WORKFLOWS['valid']['app_label'], INVALID_VERSION)

        # Loading a version should create the right database objects.
        load_workflow(WORKFLOWS['valid']['app_label'], VERSION_1)
        assert_test_dir_workflow_loaded(self)
        assert_test_dir_v1_loaded(self)
        assert_test_dir_v2_not_loaded(self)

        # Loading a second version should create the right database objects.
        load_workflow(WORKFLOWS['valid']['app_label'], VERSION_2)
        assert_test_dir_workflow_loaded(self)
        assert_test_dir_v1_loaded(self)
        assert_test_dir_v2_loaded(self)

        # Reloading a version without --force should fail.
        force_error = 'Version {} already exists'.format(VERSION_1)
        with self.assertRaisesMessage(WorkflowError, force_error):
            load_workflow(WORKFLOWS['valid']['app_label'], VERSION_1)

        # Reloading versions with --force should succeed.
        load_workflow(WORKFLOWS['valid']['app_label'], VERSION_1, force=True)
        load_workflow(WORKFLOWS['valid']['app_label'], VERSION_2, force=True)
        assert_test_dir_workflow_loaded(self)
        assert_test_dir_v1_loaded(self)
        assert_test_dir_v2_loaded(self)

    def test_load_workflow_version(self):
        """ Ensure that workflow version loading works as desired. """
        # Verify initial DB state.
        assert_test_dir_workflow_not_loaded(self)
        assert_test_dir_v1_not_loaded(self)
        assert_test_dir_v2_not_loaded(self)

        # Load V1 of the workflow.
        load_workflow(WORKFLOWS['valid']['app_label'], VERSION_1)
        workflow = Workflow.objects.get(slug='test_dir')
        assert_test_dir_workflow_loaded(self)
        assert_test_dir_v1_loaded(self)
        assert_test_dir_v2_not_loaded(self)

        # Load the JSON data for the versions.
        v1_file_path = os.path.join(WORKFLOWS['valid']['dir'],
                                    'v1/version.json')
        with open(v1_file_path, 'r') as v1_file:
            v1_data = json.load(v1_file)

        v2_file_path = os.path.join(WORKFLOWS['valid']['dir'],
                                    'v2/version.json')
        with open(v2_file_path, 'r') as v2_file:
            v2_data = json.load(v2_file)

        # Without --force, can't overwrite a version.
        # We wrap calls to load_workflow_version in transaction.atomic, because
        # the call might create corrupt database state otherwise.
        force_error = 'Version {} already exists'.format(VERSION_1)
        with self.assertRaisesMessage(WorkflowError, force_error):
            with transaction.atomic():
                load_workflow_version(v1_data, workflow)

        # Even with --force, can't overwrite a version with fewer steps
        removed_step = v1_data['steps'][-1]
        v1_data['steps'] = v1_data['steps'][:-1]
        step_remove_err_msg = (
            'Even with --force, you cannot remove steps from a workflow.'
            'Drop and recreate the database to reset, or create a new '
            'version for your workflow.')
        with self.assertRaisesMessage(WorkflowError, step_remove_err_msg):
            with transaction.atomic():
                load_workflow_version(
                    v1_data, workflow, force=True)
        v1_data['steps'].append(removed_step)

        # Even with --force, can't change a step's creation dependencies.
        step_2_create_dependencies = v1_data['steps'][1]['creation_depends_on']
        step_2_create_dependencies.append('s3')
        topology_change_err_msg = ('Even with --force, you cannot change the '
                                   'topology of a workflow. Drop and recreate '
                                   'the database to reset, or create a new '
                                   'version for your workflow.')
        with self.assertRaisesMessage(WorkflowError, topology_change_err_msg):
            with transaction.atomic():
                load_workflow_version(v1_data, workflow, force=True)
        v1_data['steps'][1]['creation_depends_on'] = (
            step_2_create_dependencies[:-1])

        # Otherwise, --force should reload versions correctly.
        with transaction.atomic():
            load_workflow_version(v1_data, workflow, force=True)
        assert_test_dir_workflow_loaded(self)
        assert_test_dir_v1_loaded(self)
        assert_test_dir_v2_not_loaded(self)

        # --force should reload even if we add a step to the workflow.
        new_step = deepcopy(v1_data['steps'][-1])
        new_step.update({
            'slug': 's4',
            'name': 'Step 4',
            'description': 'The fourth step',
        })
        v1_data['steps'].append(new_step)
        with transaction.atomic():
            load_workflow_version(v1_data, workflow, force=True)
        assert_test_dir_workflow_loaded(self)
        assert_test_dir_v1_loaded(self, extra_step=True)
        assert_test_dir_v2_not_loaded(self)
        v1_data['steps'] = v1_data['steps'][:-1]

        # New versions with bad slugs should not load correctly
        v2_step_2 = v2_data['steps'][1]
        v2_step_2_create_dependencies = v2_step_2['creation_depends_on']
        v2_step_2_create_dependencies.append('not_a_real_step')
        bad_slug_error = '{}.{} contains a non-existent slug'
        with self.assertRaisesMessage(
                WorkflowError,
                bad_slug_error.format('s2', 'creation_depends_on')):
            with transaction.atomic():
                load_workflow_version(v2_data, workflow)
        v2_step_2['creation_depends_on'] = (
            v2_step_2_create_dependencies[:-1])

        v2_step_2_submit_dependencies = v2_step_2['submission_depends_on']
        v2_step_2_submit_dependencies.append('not_a_real_step')
        with self.assertRaisesMessage(
                WorkflowError,
                bad_slug_error.format('s2', 'submission_depends_on')):
            with transaction.atomic():
                load_workflow_version(v2_data, workflow)
        v2_step_2['submission_depends_on'] = (
            v2_step_2_submit_dependencies[:-1])

        v2_step_2_certification_dependencies = v2_step_2[
            'required_certifications']
        v2_step_2_certification_dependencies.append('not_a_real_certification')
        with self.assertRaisesMessage(
                WorkflowError,
                bad_slug_error.format('s2', 'required_certifications')):
            with transaction.atomic():
                load_workflow_version(v2_data, workflow)
        v2_step_2['required_certifications'] = (
            v2_step_2_certification_dependencies[:-1])

        # Otherwise, new versions should load correctly
        with transaction.atomic():
            load_workflow_version(v2_data, workflow)
        assert_test_dir_workflow_loaded(self)
        assert_test_dir_v1_loaded(self)
        assert_test_dir_v2_loaded(self)



# Generated by Django 2.2.13 on 2021-02-11 21:57

from django.db import migrations, models
import orchestra.models.core.models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0092_autostaff_updates'),
    ]

    operations = [
        migrations.AlterField(
            model_name='todo',
            name='status',
            field=models.IntegerField(choices=[(0, 'pending'), (1, 'completed'), (2, 'declined')], default=0),
        ),
    ]




from django.test import TestCase
from unittest.mock import patch

from orchestra.models import Todo
from orchestra.tests.helpers.fixtures import WorkerFactory
from orchestra.tests.helpers.fixtures import UserFactory
from orchestra.tests.helpers.fixtures import TodoFactory
from orchestra.tests.helpers.fixtures import StepFactory
from orchestra.tests.helpers.fixtures import ProjectFactory
from orchestra.utils.common_helpers import get_update_message
from orchestra.utils.common_helpers import notify_single_todo_update


class ViewHelpersTests(TestCase):
    def setUp(self):
        super().setUp()
        self.project = ProjectFactory()
        step = StepFactory()
        self.old_title = 'Old title'
        self.new_title = 'New title'
        self.old_details = 'Old details'
        self.new_details = 'New details'
        self.old_todo = TodoFactory(
            title=self.old_title,
            details=self.old_details,
            project=self.project,
            step=step)
        self.new_todo = TodoFactory(
            title=self.new_title,
            details=self.new_details,
            status=Todo.Status.COMPLETED.value,
            project=self.project,
            step=step)
        self.sender = UserFactory()

    def test_empty_msg_if_no_updates(self):
        old_todo = self.new_todo
        msg = get_update_message(old_todo, self.new_todo, self.sender)
        expected_msg = ''
        self.assertEqual(msg, expected_msg)

    def test_fields_updated_completed_with_sender(self):
        msg = get_update_message(self.old_todo, self.new_todo, self.sender)
        expected_msg = (
            '{} has updated `{}`: marked complete, '
            'changed title and details'
        ).format(self.sender.username, self.new_todo.title)
        self.assertEqual(msg, expected_msg)

        # Reversing the change
        msg = get_update_message(self.new_todo, self.old_todo, self.sender)
        expected_msg = (
            '{} has updated `{}`: marked incomplete, '
            'changed title and details'
        ).format(self.sender.username, self.old_todo.title)
        self.assertEqual(msg, expected_msg)

    def test_fields_updated_completed_without_sender(self):
        msg = get_update_message(self.old_todo, self.new_todo)
        expected_msg = (
            '`{}` has been updated: marked complete, '
            'changed title and details'
        ).format(self.new_todo.title)
        self.assertEqual(msg, expected_msg)

        # Reversing the change
        msg = get_update_message(self.new_todo, self.old_todo)
        expected_msg = (
            '`{}` has been updated: marked incomplete, '
            'changed title and details'
        ).format(self.old_todo.title)
        self.assertEqual(msg, expected_msg)

    def test_fields_not_updated_completed(self):
        new_todo = TodoFactory(
            title=self.old_title,
            details=self.old_details,
            status=Todo.Status.COMPLETED.value)
        msg = get_update_message(self.old_todo, new_todo, self.sender)
        expected_msg = '{} has updated `{}`: marked complete'.format(
            self.sender.username, new_todo.title)
        self.assertEqual(msg, expected_msg)

        # Reversing the change
        msg = get_update_message(new_todo, self.old_todo, self.sender)
        expected_msg = '{} has updated `{}`: marked incomplete'.format(
            self.sender.username, new_todo.title)
        self.assertEqual(msg, expected_msg)

    def test_fields_not_updated_not_relevant(self):
        new_todo = TodoFactory(
            title=self.old_title,
            details=self.old_details,
            status=Todo.Status.DECLINED.value)
        msg = get_update_message(self.old_todo, new_todo, self.sender)
        expected_msg = '{} has updated `{}`: marked not relevant'.format(
            self.sender.username, new_todo.title)
        self.assertEqual(msg, expected_msg)

        # No sender
        msg = get_update_message(self.old_todo, new_todo)
        expected_msg = '`{}` has been updated: marked not relevant'.format(
            new_todo.title)
        self.assertEqual(msg, expected_msg)

        # Reversing the change
        msg = get_update_message(new_todo, self.old_todo)
        expected_msg = '`{}` has been updated: marked relevant'.format(
            new_todo.title)
        self.assertEqual(msg, expected_msg)

    def test_fields_updated(self):
        old_todo = TodoFactory(
            title=self.old_title,
            details=self.old_details)
        new_todo = TodoFactory(
            title=self.new_title,
            details=self.new_details)
        msg = get_update_message(old_todo, new_todo, self.sender)
        expected_msg = (
            '{} has updated `{}`: changed title and details').format(
            self.sender.username, new_todo.title)
        self.assertEqual(msg, expected_msg)

        # Only title has changed
        old_todo = TodoFactory(
            title=self.old_title,
            details=self.old_details)
        new_todo = TodoFactory(
            title=self.new_title,
            details=self.old_details)
        msg = get_update_message(old_todo, new_todo, self.sender)
        expected_msg = (
            '{} has updated `{}`: changed title').format(
            self.sender.username, new_todo.title)
        self.assertEqual(msg, expected_msg)

    @patch('orchestra.utils.common_helpers.message_experts_slack_group')
    def test_notify_single_todo_update(self, mock_slack):
        # Level 0 todos
        WorkerFactory(user=self.sender)
        notify_single_todo_update(
            self.sender, self.old_todo, self.new_todo)
        self.assertEqual(mock_slack.call_count, 1)

        # Level 1 todos
        parent_todo_level_1 = TodoFactory(title='Parent todo')
        old_todo = TodoFactory(
            title=self.old_title,
            parent_todo=parent_todo_level_1,
            project=self.project,
            details=self.old_details)
        new_todo = TodoFactory(
            title=self.new_title,
            parent_todo=parent_todo_level_1,
            project=self.project,
            details=self.new_details)

        notify_single_todo_update(self.sender, old_todo, new_todo)
        self.assertEqual(mock_slack.call_count, 2)

        # Level 2 todos
        parent_todo_level_2 = TodoFactory(
            title='Parent todo',
            parent_todo=parent_todo_level_1)
        old_todo = TodoFactory(
            title=self.old_title,
            parent_todo=parent_todo_level_2,
            project=self.project,
            details=self.old_details)
        new_todo = TodoFactory(
            title=self.new_title,
            parent_todo=parent_todo_level_2,
            project=self.project,
            details=self.new_details)

        notify_single_todo_update(self.sender, old_todo, new_todo)
        self.assertEqual(mock_slack.call_count, 2)



import json
import logging

from annoying.functions import get_object_or_None
from collections import defaultdict
from datetime import timedelta
from django.conf import settings
from django.urls import reverse
from django.db import transaction
from django.db.models import Q
from django.utils import timezone
from markdown2 import markdown

from orchestra.bots.errors import StaffingResponseException
from orchestra.bots.staffbot import StaffBot
from orchestra.core.errors import TaskAssignmentError
from orchestra.core.errors import TaskStatusError
from orchestra.models import CommunicationPreference
from orchestra.models import StaffBotRequest
from orchestra.models import StaffingRequestInquiry
from orchestra.models import StaffingResponse
from orchestra.models import Task
from orchestra.models import Project
from orchestra.models import TaskAssignment
from orchestra.models import Worker
from orchestra.models import WorkerAvailability
from orchestra.models import WorkerCertification
from orchestra.utils.datetime_utils import first_day_of_the_week
from orchestra.utils.time_tracking import time_entry_hours_worked
from orchestra.utils.notifications import message_experts_slack_group
from orchestra.utils.notifications import message_internal_slack_group
from orchestra.utils.task_lifecycle import assign_task
from orchestra.utils.task_lifecycle import check_worker_allowed_new_assignment
from orchestra.utils.task_lifecycle import get_role_from_counter
from orchestra.utils.task_lifecycle import is_worker_certified_for_task
from orchestra.utils.task_lifecycle import reassign_assignment

logger = logging.getLogger(__name__)


@transaction.atomic
def handle_staffing_response(worker, staffing_request_inquiry_id,
                             is_available=False):
    """
    Args:
        worker (orchestra.models.Worker):
            Worker instance that responsed to a staffing request inquiry
        staffing_request_inquiry_id (int):
            Id of a staffing_request_inquiry that is associated with a response
        is_available (boolean):
            Boolean that tells whether worker accepted an inquiry or not
    Returns:
        response (orchestra.models.StaffingResponse):
            StaffingResponse object that has been created for the worker
    """
    staffing_request_inquiry = get_object_or_None(
        StaffingRequestInquiry,
        communication_preference__worker=worker,
        id=staffing_request_inquiry_id
    )
    if staffing_request_inquiry is None:
        return None

    response = (StaffingResponse.objects
                .filter(request_inquiry=staffing_request_inquiry))
    if response.exists():
        response = response.first()
        if not is_available and response.is_winner:
            raise StaffingResponseException(
                'Cannot reject after accepting the task')
        # This update will be saved in re/assing_task if necessary.
        response.is_available = is_available

    else:
        response = StaffingResponse.objects.create(
            request_inquiry=staffing_request_inquiry,
            is_available=is_available)

    request = staffing_request_inquiry.request
    if (is_available and
            request.status != StaffBotRequest.Status.CLOSED.value):
        task_assignment = get_object_or_None(
            TaskAssignment,
            task=request.task,
            assignment_counter=request.required_role_counter
        )

        # if task assignment exists then reassign
        if task_assignment is not None:
            reassign_assignment(worker.id, task_assignment.id,
                                staffing_request_inquiry)
        # otherwise assign task
        else:
            assign_task(worker.id, request.task.id,
                        staffing_request_inquiry)

    response = (StaffingResponse.objects
                .filter(request_inquiry=staffing_request_inquiry)
                .first())
    check_responses_complete(staffing_request_inquiry.request)
    return response


def check_responses_complete(request):
    inquiries = (
        StaffingRequestInquiry.objects.filter(request=request)
    ).distinct()
    num_inquired_workers = len(
        set(inquiries.values_list(
            'communication_preference__worker__id', flat=True)
            )
    )
    responded_inquiries = inquiries.filter(
        responses__isnull=False).distinct()
    num_responded_workers = len(
        set(responded_inquiries.values_list(
            'communication_preference__worker__id', flat=True)
            )
    )

    responses = StaffingResponse.objects.filter(
        request_inquiry__request=request)
    if (num_responded_workers >= num_inquired_workers and
            not responses.filter(is_winner=True).exists()):
        request.status = StaffBotRequest.Status.CLOSED.value
        request.save()

        # notify that all workers have rejected a task
        message_experts_slack_group(
            request.task.project.slack_group_id,
            ('No worker has accepted to work on task {}'
             .format(request.task)))


def _exclude_inactive_staffbot_requests(request_queryset):
    return (request_queryset
            .exclude(status=StaffBotRequest.Status.CLOSED.value)
            .exclude(task__project__status=Project.Status.COMPLETED)
            .exclude(task__project__status=Project.Status.ABORTED)
            .exclude(task__status=Task.Status.COMPLETE)
            .exclude(task__status=Task.Status.ABORTED))


def _get_worker_allowed_hours():
    today = timezone.now().date()
    today_abbreviation = ['mon', 'tues', 'wed', 'thurs', 'fri', 'sat', 'sun'][
        today.weekday()]
    # For each Worker, their allowed hours is the smaller of the their
    # daily (desired) availability and their max hours per day.
    return {
        availability.worker.id:
        min(
            getattr(
                availability,
                'hours_available_{}'.format(today_abbreviation), 0),
            availability.worker.max_autostaff_hours_per_day)
        for availability in (
                WorkerAvailability.objects
                .filter(week=first_day_of_the_week())
                .select_related('worker'))
    }


def address_staffing_requests(
        worker_batch_size=settings.ORCHESTRA_STAFFBOT_WORKER_BATCH_SIZE,
        frequency=settings.ORCHESTRA_STAFFBOT_BATCH_FREQUENCY):
    staffbot = StaffBot()
    cutoff_datetime = timezone.now() - frequency
    requests = _exclude_inactive_staffbot_requests(
        StaffBotRequest.objects
        .filter(Q(last_inquiry_sent__isnull=True) |
                Q(last_inquiry_sent__lte=cutoff_datetime))
        .order_by('-task__project__priority', 'created_at'))
    worker_allowed_hours = _get_worker_allowed_hours()
    for request in requests:
        staff_or_send_request_inquiries(
            staffbot, request, worker_batch_size, worker_allowed_hours)


def _is_worker_assignable(worker, task, required_role):
    try:
        check_worker_allowed_new_assignment(worker)
        if (is_worker_certified_for_task(worker, task,
                                         required_role,
                                         require_staffbot_enabled=True)
                and not task.is_worker_assigned(worker)):
            return True
    except TaskStatusError:
        pass
    except TaskAssignmentError:
        pass
    return False


def _can_handle_more_work_today(worker, task, worker_allowed_hours):
    can_handle_more_hours = False
    today = timezone.now().date()
    allowed_hours = worker_allowed_hours.get(worker.id, None)
    task_hours = task.get_assignable_hours()
    if allowed_hours is not None and task_hours is not None:
        # Because we're looking at StaffingResponse objects to
        # determine assigned tasks for a worker, we don't consider
        # tasks that were assigned manually without an open StaffBot
        # request. For example, we ignore tasks staffed by directly
        # typing a worker's username into the projman interface when
        # an open StaffBot request does not exist for the task. In
        # practice, such situations are rare (new tasks tend to be
        # auto-StaffBotted), and when the situations arise, it's
        # unclear what the estimate for the assignable hours is.
        responses = StaffingResponse.objects.filter(
            request_inquiry__communication_preference__worker=worker,
            is_winner=True,
            created_at__gte=today,
            created_at__lt=today + timedelta(days=1)
        )
        # Create a dictionary to deduplicate tasks if a Worker is
        # reassigned the same task multiple times.
        hours_assigned = {
            response.request_inquiry.request.task.id:
            (response.request_inquiry.request.task.get_assignable_hours(),
             response.request_inquiry.request.task)
            for response in responses
        }
        hours_assigned = [
            (hours, task) for (hours, task) in hours_assigned.values()
            if hours is not None]
        max_tasks = settings.ORCHESTRA_MAX_AUTOSTAFF_TASKS_PER_DAY
        # To estimate how much someone worked today, we add:
        # - the number of assignable hours they were assigned today
        # - the number of hours they tracked (excluding work completed
        #   on today's newly assigned work, to avoid double-counting)
        # We do not attempt to estimate the amount of "unexpected" work,
        #   like iteration time on an old project the expert has learned
        #   about over Slack but hasn't yet logged for the day.
        sum_hours_assigned = sum(hours for (hours, task) in hours_assigned)
        sum_hours_worked = time_entry_hours_worked(
            today, worker, excluded_tasks=[
                task for (hours, task) in hours_assigned])
        can_handle_more_hours = (
            (len(hours_assigned) + 1 <= max_tasks)
            and (sum_hours_assigned
                 + sum_hours_worked
                 + task_hours <= allowed_hours))
    return can_handle_more_hours


def _attempt_to_automatically_staff(
        staffbot, request, worker_certifications, worker_allowed_hours):
    successfully_staffed = False
    required_role = get_role_from_counter(request.required_role_counter)
    attempted_workers = set()
    new_task_available_type = (
        CommunicationPreference.CommunicationType.NEW_TASK_AVAILABLE.value)
    previously_opted_in_method = (
        StaffingRequestInquiry.CommunicationMethod.PREVIOUSLY_OPTED_IN.value)
    worker = None
    for certification in worker_certifications:
        worker = certification.worker
        if worker.id in attempted_workers:
            continue
        attempted_workers.add(worker.id)
        if (_is_worker_assignable(worker, request.task, required_role)
                and _can_handle_more_work_today(
                    worker, request.task, worker_allowed_hours)):
            communication_preference = (
                CommunicationPreference.objects.get(
                    communication_type=new_task_available_type,
                    worker=worker))
            staffing_request_inquiry = StaffingRequestInquiry.objects.create(
                communication_preference=communication_preference,
                communication_method=previously_opted_in_method,
                request=request)
            handle_staffing_response(
                worker, staffing_request_inquiry.id, is_available=True)
            successfully_staffed = True
            break
    logger.info('Autostaff attempt: %s', json.dumps({
        'successfully_staffed': successfully_staffed,
        'attempted_workers': list(attempted_workers),
        'most_recent_worker': str(worker) if worker else None,
        'most_recent_worker_id': worker.id if worker else None,
        'staffing_request': str(request),
        'staffing_request_id': request.id,
        'task': str(request.task),
        'task_id': request.task.id
    }))
    return successfully_staffed


def _send_request_inquiries(staffbot, request, worker_batch_size,
                            worker_certifications):
    inquiries_sent = 0
    required_role = get_role_from_counter(request.required_role_counter)
    contacted_workers = set()
    for certification in worker_certifications:
        worker = certification.worker
        if worker.id in contacted_workers:
            continue
        contacted_workers.add(worker.id)
        if _is_worker_assignable(worker, request.task, required_role):
            staffbot.send_task_to_worker(worker, request)
            inquiries_sent += 1
        if inquiries_sent >= worker_batch_size:
            break

    # check whether all inquiries have been sent out.
    if inquiries_sent < worker_batch_size:
        message_experts_slack_group(
            request.task.project.slack_group_id,
            ('All staffing requests for task {} have been sent!'
             .format(request.task)))
        request.status = StaffBotRequest.Status.DONE_SENDING_INQUIRIES.value
    request.last_inquiry_sent = timezone.now()
    request.save()


def staff_or_send_request_inquiries(
        staffbot, request, worker_batch_size, worker_allowed_hours):
    # Get Workers that haven't already received an inquiry.
    workers_with_inquiries = (StaffingRequestInquiry.objects.filter(
        request=request).distinct().values_list(
            'communication_preference__worker__id', flat=True))
    required_role = get_role_from_counter(request.required_role_counter)

    # Sort Worker Certifications by their staffing priority first,
    # and then randomly within competing staffing priorities.
    worker_certifications = (
        WorkerCertification
        .objects
        .filter(role=required_role,
                task_class=WorkerCertification.TaskClass.REAL,
                certification__in=(request.task
                                   .step.required_certifications.all()))
        .order_by('-staffing_priority', '?'))
    available_worker_certifications = (
        worker_certifications
        .filter(worker__availabilities__week=first_day_of_the_week()))
    uninquired_worker_certifications = (
        worker_certifications
        .exclude(worker__id__in=workers_with_inquiries))
    successfully_staffed = _attempt_to_automatically_staff(
        staffbot, request, available_worker_certifications,
        worker_allowed_hours)
    sending_inquiries = StaffBotRequest.Status.SENDING_INQUIRIES.value
    # We consider StaffBotRequests that are done sending inquiries
    # when auto-staffing, since it's possible for a worker to have new
    # auto-staffing availability for a request that has already been
    # sent to them (e.g., the day after receiving a request, their new
    # availability kicks in). Once we reach the branch below, we only
    # want to send new request inquiries for requests that aren't
    # `DONE_SENDING_INQUIRIES`.
    if ((not successfully_staffed)
            and (request.status == sending_inquiries)):
        _send_request_inquiries(staffbot,
                                request,
                                worker_batch_size,
                                uninquired_worker_certifications)


def get_available_requests(worker):
    # We want to show a worker only requests for which there is no
    # winner or for which they have not already replied.
    worker_provided_responses = StaffingResponse.objects.filter(
        request_inquiry__communication_preference__worker=worker)
    remaining_requests = _exclude_inactive_staffbot_requests(
        StaffBotRequest.objects
        .filter(inquiries__communication_preference__worker=worker)
        .exclude(inquiries__responses__in=worker_provided_responses)
        .distinct())
    inquiries = (
        StaffingRequestInquiry.objects
        .filter(request__in=remaining_requests)
        .filter(communication_preference__worker=worker)
        .order_by('-request__task__project__priority', 'request__created_at'))
    # Because we might send multiple request inquiries to the same
    # worker for the same request (e.g., email and slack), we
    # deduplicate the inquiries so that we will return at most one
    # inquiry's worth of content here.
    request_ids = set()
    contexts = []
    staffbot = StaffBot()
    for inquiry in inquiries:
        if inquiry.request.id in request_ids:
            continue
        request_ids.add(inquiry.request.id)
        metadata = staffbot.get_staffing_request_metadata(inquiry)
        metadata['detailed_description'] = markdown(
            metadata['detailed_description'], extras=['target-blank-links'])
        metadata['reject_url'] += '?next={}'.format(
            reverse('orchestra:communication:available_staffing_requests'))
        contexts.append(metadata)
    return contexts


def warn_staffing_team_about_unstaffed_tasks():
    max_unstaffed_datetime = (
        timezone.now() - settings.ORCHESTRA_STAFFBOT_STAFFING_MIN_TIME)

    # Get all requests without winners
    task_values = (
        Task.objects.all()
        .filter(start_datetime__lt=max_unstaffed_datetime)
        .exclude(staffing_requests__inquiries__responses__is_winner=True)
        .exclude(status=Task.Status.COMPLETE)
        .exclude(status=Task.Status.ABORTED)
        .exclude(staffing_requests__isnull=True)
        .exclude(staffing_requests__inquiries__isnull=True)
        .order_by('-start_datetime')
        .values('staffing_requests__required_role_counter', 'id'))

    requests_to_notify = defaultdict(list)
    for task_value in task_values:
        required_role_counter = task_value[
            'staffing_requests__required_role_counter']
        request = (
            StaffBotRequest.objects.filter(
                task=task_value['id'],
                required_role_counter=required_role_counter)
            .order_by('-created_at'))[0]

        if request.created_at < max_unstaffed_datetime:
            requests_to_notify[request.task.step].append(request)

    for step, requests in requests_to_notify.items():
        message = '\n'.join([
            ('No winner request for task {} - {}! Created at {}.'
                .format(request.task.id, request.task, request.created_at))
            for request in requests])
        message_internal_slack_group(
            settings.ORCHESTRA_STAFFBOT_STAFFING_GROUP_ID, message)


def remind_workers_about_available_tasks():
    staffbot = StaffBot()
    workers = Worker.objects.all()
    for worker in workers:
        requests = get_available_requests(worker)
        # TODO(kkamalov): send out reminder only if last request was sent
        # at least ORCHESTRA_STAFFBOT_MIN_FOLLOWUP_TIME ago
        if len(requests):
            staffbot.send_worker_tasks_available_reminder(worker)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0020_auto_20151028_1559'),
        ('orchestra', '0020_auto_20151022_1553'),
    ]

    operations = [
    ]



from datetime import timedelta
from django.utils import timezone


def first_day_of_the_week(relative_to_date=None):
    if relative_to_date is None:
        relative_to_date = timezone.now()
    return (relative_to_date -
            timedelta(days=relative_to_date.weekday())).date()



from django import forms
from django.contrib.auth import get_user_model
from phonenumber_field.formfields import PhoneNumberField
from phonenumber_field.widgets import PhoneNumberPrefixWidget

from orchestra.accounts.bitformfield import BitFormField
from orchestra.communication.slack import get_slack_user_id
from orchestra.models import CommunicationPreference
from orchestra.models import Worker
from orchestra.models import WorkerAvailability


UserModel = get_user_model()


class UserForm(forms.ModelForm):
    first_name = forms.CharField()
    last_name = forms.CharField()

    class Meta:
        model = UserModel
        fields = ('first_name', 'last_name')
        # TODO(joshblum): support change email
        exclude = ('is_staff', 'is_active', 'date_joined', 'username', 'email')


class DailyAvailabilityField(forms.FloatField):
    def __init__(self, *args, **kwargs):
        max_value = 24  # only 24 hours in a day!
        super().__init__(*args, min_value=0, max_value=max_value, **kwargs)


class WorkerAvailabilityForm(forms.ModelForm):
    class Meta:
        model = WorkerAvailability
        fields = (
            'hours_available_mon',
            'hours_available_tues',
            'hours_available_wed',
            'hours_available_thurs',
            'hours_available_fri',
            'hours_available_sat',
            'hours_available_sun',
        )
        field_classes = {field: DailyAvailabilityField for field in fields}


class WorkerForm(forms.ModelForm):
    slack_username = forms.CharField()
    phone = PhoneNumberField(
        widget=PhoneNumberPrefixWidget(initial='US'))

    class Meta:
        model = Worker
        fields = ('slack_username', 'phone')
        exclude = ('user', 'start_datetime', 'slack_user_id')

    def clean_slack_username(self):
        slack_username = self.cleaned_data.get('slack_username')
        if slack_username:
            slack_user_id = get_slack_user_id(slack_username)
            if slack_user_id is None:
                raise forms.ValidationError(
                    'Incorrect slack username provided')
            self.cleaned_data['slack_user_id'] = slack_user_id
        return self.cleaned_data['slack_username']

    def save(self, commit=True):
        instance = super().save(commit=False)
        instance.slack_user_id = self.cleaned_data.get('slack_user_id')
        if commit:
            instance.save()
        return instance


class CommunicationPreferenceForm(forms.ModelForm):
    methods = BitFormField()

    class Meta:
        model = CommunicationPreference
        fields = ('methods',)
        exclude = ('worker', 'is_deleted',
                   'created_at')



from django.contrib.auth.decorators import login_required
from django.http import Http404
from django.http import HttpResponseRedirect
from django.shortcuts import render

from orchestra.bots.errors import StaffingResponseException
from orchestra.communication.staffing import get_available_requests
from orchestra.communication.staffing import handle_staffing_response
from orchestra.models import Worker


@login_required
def accept_staffing_request_inquiry(request,
                                    staffing_request_inquiry_id):
    worker = Worker.objects.get(user=request.user)
    response = handle_staffing_response(
        worker, staffing_request_inquiry_id, is_available=True)

    if response is None:
        raise Http404
    return render(request, 'communication/staffing_request_accepted.html',
                  {
                      'response': response,
                  })


@login_required
def reject_staffing_request_inquiry(request,
                                    staffing_request_inquiry_id):
    worker = Worker.objects.get(user=request.user)
    try:
        response = handle_staffing_response(
            worker, staffing_request_inquiry_id, is_available=False)
    except StaffingResponseException:
        return render(request,
                      'communication/staffing_response_not_permitted.html',
                      {})
    if response is None:
        raise Http404

    next_path = request.GET.get('next')
    if next_path:
        return HttpResponseRedirect(next_path)
    else:
        return render(request, 'communication/staffing_request_rejected.html',
                      {})


@login_required
def available_staffing_requests(request):
    worker = Worker.objects.get(user=request.user)
    return render(request, 'communication/available_staffing_requests.html',
                  {
                      'requests': get_available_requests(worker),
                  })



def get_default_assignment_policy(is_human):
    """ Return the default assignment policy.

    Args:
        is_human (bool):
            Indicates whether the policy is for a human or machine.

    Returns:
        default_policy (dict):
            A dictionary specifying the assignment policy.
    """
    return {} if not is_human else {
        'policy_function': {
            'entry_level': {
                'path': 'orchestra.assignment_policies.anyone_certified'
            },
            'reviewer': {
                'path': 'orchestra.assignment_policies.anyone_certified'
            }
        },
    }


def get_default_creation_policy():
    return {
        'policy_function': {
            'path': 'orchestra.creation_policies.always_create',
        }
    }


def get_default_review_policy(is_human):
    """ Return the default review policy.

    Args:
        is_human (bool):
            Indicates whether the policy is for a human or machine.

    Returns:
        default_policy (dict):
            A dictionary specifying the assignment policy.
    """
    return {} if not is_human else {
        'policy': 'sampled_review',
        'rate': 1,
        'max_reviews': 1
    }



from rest_framework import permissions

from orchestra.models import Worker
from orchestra.models import Todo
from orchestra.models import TodoQA


class IsAssociatedWithTodosProject(permissions.BasePermission):
    """
    Ensures that a user's worker is accoiated with the todo's project.
    """

    def has_object_permission(self, request, view, obj):
        worker = Worker.objects.get(user=request.user)
        if isinstance(obj, Todo):
            project = obj.project
        elif isinstance(obj, TodoQA):
            project = obj.todo.project
        else:
            project = None
        return (
            project and
            (worker.is_project_admin() or
             worker.assignments.filter(task__project=project).exists()))


class IsAssociatedWithProject(permissions.BasePermission):
    """
    Ensures that a user's worker is associated with the request's
    `project`.

    """

    def has_permission(self, request, view):
        """
        We pass project_id as a payload in cases when the request
        is either POST, PUT or PATCH. It can be passed via query param
        not only in a GET request, but also in the requests listed above
        (when applying a filter).
        """
        worker = Worker.objects.get(user=request.user)
        if worker.is_project_admin():
            return True
        todo_id = request.data.get('todo')
        if todo_id is None:
            todo_id = view.kwargs.get('pk')
        project_id = request.data.get(
            'project') or request.data.get('project__id')
        if project_id is None:
            project_id = request.query_params.get(
                'project') or request.query_params.get('project__id')
        if project_id is None and todo_id is not None:
            project_id = Todo.objects.get(id=todo_id).project.id
        return worker.assignments.filter(task__project__id=project_id).exists()


class IsAssociatedWithTask(permissions.BasePermission):
    """
    Ensures that a user's worker is associated with the request's
    `task`.

    """

    def has_permission(self, request, view):
        worker = Worker.objects.get(user=request.user)
        if worker.is_project_admin():
            return True
        if request.method == 'GET':
            task_id = request.query_params.get('task')
            return worker.assignments.filter(task=task_id).exists()
        return False



# Generated by Django 2.2.13 on 2020-09-20 15:59

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import orchestra.models.core.mixins
import orchestra.utils.models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('orchestra', '0088_change_related_name_todos_old'),
    ]

    operations = [
        migrations.CreateModel(
            name='TodoListTemplateImportRecord',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now)),
                ('is_deleted', models.BooleanField(default=False)),
                ('import_url', models.URLField(blank=True, null=True)),
                ('importer', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL)),
                ('todo_list_template', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='orchestra.TodoListTemplate')),
            ],
            bases=(orchestra.models.core.mixins.TodoListTemplateImportRecordMixin, orchestra.utils.models.DeleteMixin, models.Model),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-12 23:06
from __future__ import unicode_literals

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0033_alter_worker_and_assignment_on_timeentry'),
    ]

    operations = [
        migrations.AddField(
            model_name='timeentry',
            name='is_deleted',
            field=models.BooleanField(default=False),
        ),
    ]



from orchestra.accounts.signals import orchestra_user_registered
from orchestra.models import CommunicationPreference
from orchestra.tests.helpers import OrchestraModelTestCase
from orchestra.tests.helpers.fixtures import CommunicationPreferenceFactory
from orchestra.tests.helpers.fixtures import StaffBotRequestFactory
from orchestra.tests.helpers.fixtures import StaffingRequestInquiryFactory
from orchestra.tests.helpers.fixtures import StaffingResponseFactory
from orchestra.tests.helpers.fixtures import UserFactory


class CommunicationPreferenceTestCase(OrchestraModelTestCase):
    __test__ = True
    model = CommunicationPreferenceFactory

    def setUp(self):
        # Create a user

        self.comm_pref = CommunicationPreferenceFactory()

    def test_can_slack(self):
        """
            Assert that the object respects the slack bit field settings.
        """
        self.assertTrue(self.comm_pref.can_slack())

        self.comm_pref.methods.slack = False
        self.assertFalse(self.comm_pref.can_slack())

    def test_can_email(self):
        """
            Assert that the object respects the email bit field settings.
        """
        self.assertTrue(self.comm_pref.can_email())

        self.comm_pref.methods.email = False
        self.assertFalse(self.comm_pref.can_email())

    def test_communication_preference_autocreation(self):
        """
            When new users register, CommunicationPreferences should be
            automatically created .
        """

        # Create a user
        user = UserFactory(username='test_registration_user',
                           password='test',
                           email='test_registration_user@test.com')

        # There should be no preferences yet
        self.assertFalse(CommunicationPreference.objects.filter(
            worker__user=user).exists())

        # Fake registering the user
        orchestra_user_registered.send(
            sender=self.__class__, user=user, request=None)

        # Expect the worker object to be created
        self.assertTrue(
            CommunicationPreference.objects.filter(worker__user=user).exists(),
            'CommunicationPreference not autocreated on User registration'
        )

        comm_prefs = CommunicationPreference.objects.filter(worker__user=user)
        self.assertEqual(comm_prefs.count(), len(
            CommunicationPreference.CommunicationType.choices()))

        for comm_pref in comm_prefs:
            for label, flag in comm_pref.methods.iteritems():
                self.assertTrue(flag)


class StaffBotRequestTestCase(OrchestraModelTestCase):
    __test__ = True
    model = StaffBotRequestFactory


class StaffingRequestInquiryTestCase(OrchestraModelTestCase):
    __test__ = True
    model = StaffingRequestInquiryFactory


class StaffingResponseTestCase(OrchestraModelTestCase):
    __test__ = True
    model = StaffingResponseFactory



from datetime import timedelta

from django.db import transaction
from django.db.models import Sum
from django.utils import timezone

from orchestra.core.errors import TimerError
from orchestra.models import TaskAssignment
from orchestra.models import TaskTimer
from orchestra.models import TimeEntry


def get_timer_object(worker):
    """
    Returns TaskTimer object associated with worker, creates one if does not
    exist.
    """
    try:
        timer = worker.timer
    except TaskTimer.DoesNotExist:
        timer = TaskTimer.objects.create(worker=worker)
    return timer


def _reset_timer(timer):
    timer.assignment = None
    timer.start_time = None
    timer.stop_time = None
    timer.description = None
    timer.save()


def start_timer(worker, assignment_id=None):
    """
    Start timer for worker, and optionally task assignment id

    Args:
        worker (orchestra.models.Worker):
            The specified worker object.
        assignment_id (int): optional
            The ID of the task assignment to be associated with the time entry.

    Returns:
        timer (orchestra.models.TaskTimer)
            newly created timer object

    Raises:
        orchestra.models.TaskAssignment.DoesNotExist:
            The specified worker is not assigned to the specified task.
        orchestra.core.errors.TimerError:
            Timer has already started.
    """
    timer = get_timer_object(worker)

    # Attach assignment if provided.
    if assignment_id:
        assignment = TaskAssignment.objects.get(id=assignment_id,
                                                worker=worker)
        timer.assignment = assignment

    if timer.start_time is not None:
        raise TimerError('Timer has already started')

    timer.start_time = timezone.now()
    timer.save()
    return timer


@transaction.atomic
def stop_timer(worker):
    """
    Stops timer for worker and creates a TimeEntry object.

    Args:
        worker (orchestra.models.Worker):
            The specified worker object.
        task_id (int):
            The ID of the task to be associated with the time entry.

    Returns:
        time_entry (orchestra.models.TimeEntry)
            newly created time entry from timer duration

    Raises:
        orchestra.core.errors.TimerError:
            Timer has not started.
    """
    timer = get_timer_object(worker)
    if timer.start_time is None:
        raise TimerError('Timer not started')
    timer.stop_time = timezone.now()
    timer.save()

    # Create TimeEntry object for timer.
    time_entry = TimeEntry.objects.create(
        worker=worker,
        date=timer.start_time.date(),
        time_worked=timer.stop_time - timer.start_time,
        assignment=timer.assignment,
        timer_start_time=timer.start_time,
        timer_stop_time=timer.stop_time,
        description=timer.description)

    # Reset timer
    _reset_timer(timer)

    return time_entry


def update_timer(worker, description, assignment_id):
    timer = get_timer_object(worker)
    timer.description = description
    if assignment_id is None:
        timer.assignment = None
    else:
        timer.assignment = TaskAssignment.objects.get(
            id=assignment_id, worker=worker)
    timer.save()


def get_timer_current_duration(worker):
    """
    Returns current time interval logged by timer for a worker.

    Args:
        worker (orchestra.models.Worker):
            The specified worker object.

    Returns:
        (datetime.timedelta)
            time interval since start of timer
    """
    timer = get_timer_object(worker)
    if timer.start_time is None:
        return None
    return timezone.now() - timer.start_time


def time_entry_hours_worked(today, worker, excluded_tasks=None):
    if excluded_tasks is None:
        excluded_tasks = []
    aggregate_time = (
        TimeEntry.objects
        .filter(date=today)
        .filter(worker=worker)
        .exclude(assignment__task__in=excluded_tasks)
        .aggregate(sum_time_worked=Sum('time_worked')))
    total_duration = aggregate_time['sum_time_worked'] or timedelta(seconds=0)
    return total_duration.total_seconds() / 3600.0



import logging
from collections import defaultdict

from orchestra.models import Project
from orchestra.models import Step
from orchestra.models import WorkflowVersion
from orchestra.project_api.serializers import ProjectSerializer
from orchestra.project_api.serializers import TaskSerializer

logger = logging.getLogger(__name__)


class MalformedDependencyException(Exception):
    pass


def get_project_information(project_ids):
    """
    output format:
    { project_id: {
        'project': serialized_project_data,
        'tasks': project_tasks_list,
        'steps': project_workflow_steps
        },
        ...
    }
    Example:
    {
        123: {
            'project': {},
            'tasks': [],
            'steps': []
        },
        ...
    }
    """
    projects = Project.objects.select_related(
        'workflow_version__workflow'
        ).filter(id__in=project_ids).prefetch_related('tasks')
    projects_dict = defaultdict(dict)
    for project in projects:
        project_id = project.id
        workflow_version = project.workflow_version
        workflow = workflow_version.workflow

        projects_dict[project_id]['project'] = ProjectSerializer(
            project).data
        projects_dict[project_id]['steps'] = get_workflow_steps(
            workflow.slug, workflow_version.slug)
        tasks = defaultdict(dict)
        for task in project.tasks.all():
            tasks[project.id][task.step.slug] = TaskSerializer(task).data
        projects_dict[project_id]['tasks'] = tasks.get(project_id, {})
    return projects_dict


def get_workflow_steps(workflow_slug, version_slug):
    """Get a sorted list of steps for a project.

    Arguments:
        workflow_slug (str):
            Unique identifier of the workflow to get steps for.
        version_slug (str):
            identifier for the version of the workflow to get steps for.

    Returns:
        workflow_steps (str):
            A list of (step_slug, step_description) tuples topologically sorted
            so that earlier steps are prerequisites for later ones.
    """
    workflow_version = WorkflowVersion.objects.get(
        slug=version_slug,
        workflow__slug=workflow_slug)

    # Build a directed graph of the step dependencies
    # TODO(dhaas): make DB access more efficient.
    graph = {}
    for step in workflow_version.steps.all():
        graph[step.slug] = [dependency.slug for dependency
                            in step.creation_depends_on.all()]
    return _traverse_step_graph(graph, workflow_version)


def _traverse_step_graph(graph, workflow_version):
    queue = []
    for key, value in graph.items():
        if value == []:
            queue.append(key)

    # TODO(derek): prevent the MalformedDependencyExceptions from being
    # possible by baking protection into the Workflow/Step classes
    if not len(queue):
        raise MalformedDependencyException("All %s workflow steps have "
                                           "dependencies. There is no start "
                                           "point." % workflow_version.slug)

    # Build the steps list in order using a breadth-first-like traversal of the
    # step dependency graph
    steps = []
    already_added = set()
    while len(queue):
        current_node = queue.pop(0)

        if current_node in already_added:
            continue

        # TODO(dhaas): make DB access more efficient (one option: cache steps
        # in the graph).
        already_added.add(current_node)
        current_step = Step.objects.get(
            workflow_version=workflow_version,
            slug=current_node)
        steps.append({
                      'id': current_step.id,
                      'slug': current_node,
                      'description': current_step.description,
                      'is_human': current_step.is_human,
                      'name': current_step.name})

        for key, dependencies in graph.items():
            if (current_node in dependencies and
                    key not in already_added):
                queue.append(key)

    return steps



"""
This is the Orchestra Project Python API.

TODO(marcua): Move this file to its own pip/github project.
"""

import json
import logging
from datetime import datetime
from time import mktime
from wsgiref.handlers import format_date_time

import requests
from django.conf import settings
from httpsig.requests_auth import HTTPSignatureAuth
from orchestra.utils.convert_key_to_int import convert_key_to_int

logger = logging.getLogger(__name__)

_httpsig_auth = HTTPSignatureAuth(key_id=settings.ORCHESTRA_PROJECT_API_KEY,
                                  secret=settings.ORCHESTRA_PROJECT_API_SECRET,
                                  algorithm='hmac-sha256')
_api_root_url = '{}/orchestra/api/project/'.format(settings.ORCHESTRA_URL)


class OrchestraError(Exception):
    pass


def _make_api_request(method, endpoint, query_params='', *args, **kwargs):
    func = getattr(requests, method)
    # Adding 'date' header as per
    # https://github.com/zzsnzmn/py-http-signature/blob/e2e2c753db7da45fab4b215d84e8d490bd708833/http_signature/sign.py#L155  # noqa
    headers = {'date': format_date_time(mktime(datetime.now().timetuple())),
               'X-Api-Version': '~6.5',
               'X-Api-Key': settings.ORCHESTRA_PROJECT_API_KEY}
    headers.update(kwargs.pop('headers', {}))
    all_kwargs = {'auth': _httpsig_auth, 'headers': headers}
    all_kwargs.update(kwargs)
    url = '{}{}/{}'.format(_api_root_url, endpoint, query_params)
    response = func(url, *args, **all_kwargs)
    if response.status_code != 200 and response.status_code != 201:
        raise OrchestraError(response.text)
    return response


def get_workflow_types():
    response = _make_api_request('get', 'workflow_types')
    return json.loads(response.text)['workflows']


def get_project_details_url(project_id):
    data = {'project_id': project_id}
    response = _make_api_request('post', 'project_details_url',
                                 data=json.dumps(data))
    url = json.loads(response.text)['project_details_url']
    return url


def create_orchestra_project(client,
                             workflow_slug,
                             workflow_version_slug,
                             description,
                             priority,
                             project_data,
                             task_class):
    data = {
        'workflow_slug': workflow_slug,
        'workflow_version_slug': workflow_version_slug,
        'description': description,
        'priority': priority,
        'project_data': project_data,
        'task_class': task_class
    }
    response = _make_api_request('post', 'create_project',
                                 data=json.dumps(data))
    project_id = json.loads(response.text)['project_id']
    return project_id


def get_project_information(project_ids):
    data = {
        'project_ids': project_ids
    }
    response = _make_api_request('post', 'project_information',
                                 data=json.dumps(data))
    return json.loads(response.text, object_hook=convert_key_to_int)


def create_todos(todos):
    response = _make_api_request('post', 'todo-api',
                                 headers={'Content-type': 'application/json'},
                                 data=json.dumps(todos))
    return json.loads(response.text)


def build_url_params(project_id, step_slug, **filters):
    project_param = 'project__id={}'.format(project_id) if project_id else ''
    step_slug_param = '&step__slug={}'.format(
        step_slug) if step_slug is not None else ''
    additional_filters = '&q={}'.format(
        json.dumps(filters)) if filters else ''
    query_params = '?{}{}{}'.format(
        project_param, step_slug_param, additional_filters)
    return query_params


def get_todos(project_id, step_slug=None, **filters):
    """
    project_id: int
    step_slug: str
    filters: dict. Example: {'some_fk_field__slug': 'cool_slug'}
    See GenericTodoViewset and QueryParamsFilterBackend
    to get more info on filtering.
    """
    if project_id is None and 'id__in' not in filters:
        raise OrchestraError('project_id is required')
    query_params = build_url_params(project_id, step_slug, **filters)
    response = _make_api_request('get', 'todo-api', query_params)
    return json.loads(response.text)


def update_todos(updated_todos):
    response = _make_api_request('patch', 'todo-api',
                                 headers={'Content-type': 'application/json'},
                                 data=json.dumps(updated_todos))
    return json.loads(response.text)


def delete_todos(todo_ids):
    response = _make_api_request('delete', 'todo-api',
                                 data=json.dumps(todo_ids))
    return json.loads(response.text)


def assign_worker_to_task(worker_id, task_id):
    data = {
        'worker_id': worker_id,
        'task_id': task_id,
    }
    response = _make_api_request('post', 'assign_worker_to_task',
                                 data=json.dumps(data))
    return json.loads(response.text)


def message_project_team(project_id, message):
    data = {
        'project_id': project_id,
        'message': message
    }
    response = _make_api_request('post', 'message_project_team',
                                 data=json.dumps(data))
    return json.loads(response.text)


def get_todo_templates():
    response = _make_api_request('get', 'todo_templates')
    return json.loads(response.text)


def create_todos_from_template(todolist_template_slug, project_id,
                               step_slug, additional_data):
    data = {
        'todolist_template_slug': todolist_template_slug,
        'project_id': project_id,
        'step_slug': step_slug,
        'additional_data': additional_data,
    }
    response = _make_api_request('post', 'create_todos_from_template',
                                 data=json.dumps(data))
    return json.loads(response.text)



from django.contrib.auth.models import User

from orchestra.models import Worker


def load(workflow_version):
    """ Loads a single worker for running the simple workflow. """
    user, created = User.objects.update_or_create(
        username='demo',
        defaults={
            'first_name': 'Demo',
            'last_name': 'User',
            'is_active': True,
            'is_superuser': False,
            'is_staff': False,
            'email': 'noreply-simple1@example.org'
        }
    )
    user.set_password('demo')
    user.save()
    if created:
        Worker.objects.create(user=user)



from django.conf import settings
from django.views.decorators.csrf import csrf_exempt
from django.contrib.auth.decorators import login_required

from functools import wraps
from jsonview.decorators import json_view
from rest_framework.decorators import api_view
from rest_framework.decorators import authentication_classes
from rest_framework.decorators import permission_classes


def decorator_with_arguments(f):
    """
        A decorator decorator, allowing the decorator to be used as:
        @decorator(with, arguments, and=kwargs)
        or
        @decorator
        http://stackoverflow.com/questions/653368/how-to-create-a-python-decorator-that-can-be-used-either-with-or-without-paramet
    """
    @wraps(f)
    def new_dec(*args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            # actual decorated function
            return f(args[0])
        else:
            # decorator arguments
            return lambda realf: f(realf, *args, **kwargs)

    return new_dec


@decorator_with_arguments
def api_exception_logger(func, logger):
    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logger.exception('An API exception occurred')
            raise e
    return func_wrapper


def api_endpoint(methods, permissions, logger, auths=()):
    def programmatic_api_endpoint_decorator(func):
        @csrf_exempt
        @api_view(methods)
        @authentication_classes(auths)
        @permission_classes(permissions)
        @json_view
        @api_exception_logger(logger)
        def func_wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return func_wrapper

    def api_endpoint_decorator(func):
        @csrf_exempt
        @api_view(methods)
        @login_required
        @permission_classes(permissions)
        @json_view
        @api_exception_logger(logger)
        def func_wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return func_wrapper

    if len(auths) > 0:
        return programmatic_api_endpoint_decorator
    else:
        return api_endpoint_decorator


def run_if(*args):
    """
    Decorator prevents function run if required feature flag is False and
    raises exception if it doesn't exist.
    """
    feature_flags = args

    def require_settings_decorator(func):
        def wrapper(*args, **kwargs):
            for flag in feature_flags:
                flag_value = getattr(settings, flag, None)
                if not isinstance(flag_value, bool):
                    raise Exception('Setting {} must be either True or '
                                    'False.'.format(flag))
                elif not flag_value:
                    # Don't run the function if the flag is false.
                    return
            return func(*args, **kwargs)
        return wrapper
    return require_settings_decorator



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-29 06:44
from __future__ import unicode_literals

from enum import Enum

from django.db import migrations

################ CommunicationType Migration Boilerplate Start ################
# NOTE(joshblum): Future migrations that add a CommunicationType should copy
# paste all of this boilerplate code, so you just have to specify the correct
# COMMUNICATION_METHODS you want to be configurable.


class CommunicationMethods(object):
    # Copy pasted from models since migrations does not have access
    SLACK = 'slack'
    EMAIL = 'email'


COMMUNICATION_METHODS = (
    (CommunicationMethods.SLACK, 'Slack'),
    (CommunicationMethods.EMAIL, 'Email'),
)


class CommunicationType(Enum):
    TASK_STATUS_CHANGE = 0
    NEW_TASK_AVAILABLE = 1


def _add_communication_type(apps, schema_editor, communication_type):
    Worker = apps.get_model('orchestra', 'Worker')
    CommunicationPreference = apps.get_model('orchestra',
                                             'CommunicationPreference')
    for worker in Worker.objects.all():
        communication_preference, created = (
            CommunicationPreference.objects.get_or_create(
                worker=worker,
                communication_type=communication_type))

        # By default set both Slack and Email notifications to True
        if created:
            communication_preference.methods.slack = True
            communication_preference.methods.email = True
        communication_preference.save()
################# CommunicationType Migration Boilerplate Start ##########


def add_communication_type(apps, schema_editor):
    # Change this in future migrations to be the new CommunicationType you've
    # added.
    communication_type = (
        CommunicationType.NEW_TASK_AVAILABLE.value)
    _add_communication_type(apps, schema_editor, communication_type)


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0055_auto_20160429_0823'),
    ]

    operations = [
        migrations.RunPython(add_communication_type),  # manually-reviewed
    ]



# Generated by Django 2.2.13 on 2021-04-19 17:47

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0093_todo_status_default'),
    ]

    operations = [
        migrations.RenameField(
            model_name='project',
            old_name='team_messages_url',
            new_name='scratchpad_url',
        ),
    ]



# -*- coding: utf-8 -*-
# Manually written
from __future__ import unicode_literals

import logging

from django.db import migrations
from django.db import models

logger = logging.getLogger(__name__)


def fill_in_worker_on_time_entries(apps, schema_editor):
    TimeEntry = apps.get_model('orchestra', 'TimeEntry')

    for time_entry in TimeEntry.objects.all():
        if time_entry.assignment.worker is None:
            print('WARN: TaskAssignment {} has no worker'.format(
                time_entry.assignment.id))
        time_entry.worker = time_entry.assignment.worker
        time_entry.save()


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0030_tasktimer_and_timer_fields_on_timeentry'),
    ]

    operations = [
        migrations.RunPython(fill_in_worker_on_time_entries),  # manually-reviewed
    ]



from django.db import transaction
import logging
import operator
from pydoc import locate

from orchestra.models import TodoListTemplate
from orchestra.models import Todo
from orchestra.models import Project
from orchestra.utils.common_helpers import get_step_by_project_id_and_step_slug

logger = logging.getLogger(__name__)

OPERATORS = {
    '<': operator.lt,
    '<=': operator.le,
    '==': operator.eq,
    '!=': operator.ne,
    '>=': operator.ge,
    '>': operator.gt
}


@transaction.atomic
def add_todolist_template(todolist_template_slug, project_id,
                          step_slug, additional_data=None):
    todolist_template = TodoListTemplate.objects.get(
        slug=todolist_template_slug)

    project = Project.objects.get(id=project_id)
    step = get_step_by_project_id_and_step_slug(project_id, step_slug)
    template_todos = todolist_template.todos.get('items', [])
    additional_data = additional_data if additional_data else {}
    root_todo = Todo(
        project=project,
        step=step,
        title=todolist_template.name,
        template=todolist_template,
        additional_data=additional_data,
        status=Todo.Status.PENDING.value
    )
    root_todo.save()

    cond_props = {}
    path = todolist_template.conditional_property_function.get(
        'path', None)
    if path:
        try:
            get_cond_props = locate(path)
            cond_props = get_cond_props(project)
        except Exception:
            logger.exception('Invalid conditional function path.')
    for template_todo in template_todos:
        _add_template_todo(
            template_todo, todolist_template,
            root_todo, project, step, cond_props, additional_data)


def _to_exclude(props, conditions):
    """
    The conditions is it a list of conditions that get ORed together,
    with predicates in each dictionary getting ANDed.
    """
    any_condition_true = False

    for condition in conditions:
        all_props_true = len(condition) > 0
        for prop, predicate in condition.items():
            current_value = props.get(prop)
            compared_to_value = predicate['value']
            compare = OPERATORS[predicate['operator']]
            all_props_true = (
                all_props_true and
                compare(current_value, compared_to_value))
        any_condition_true = any_condition_true or all_props_true

    return any_condition_true


def _add_template_todo(
        template_todo, todolist_template,
        parent_todo, project, step, conditional_props,
        additional_data):
    remove = _to_exclude(conditional_props, template_todo.get('remove_if', []))
    if not remove:
        if parent_todo.status == Todo.Status.DECLINED.value:
            to_skip = True
        else:
            to_skip = _to_exclude(
                conditional_props, template_todo.get('skip_if', []))

        if to_skip:
            status = Todo.Status.DECLINED.value
        else:
            status = Todo.Status.PENDING.value
        todo = Todo(
            project=project,
            step=step,
            title=template_todo['description'],
            slug=template_todo.get('slug'),
            template=todolist_template,
            parent_todo=parent_todo,
            status=status,
            additional_data=additional_data,
            required=template_todo.get('required', False)
        )
        todo.save()
        for template_todo_item in template_todo.get('items', []):
            _add_template_todo(
                template_todo_item, todolist_template, todo,
                project, step, conditional_props, additional_data)



from urllib.parse import urljoin

from django.conf import settings
from django.urls import reverse

from orchestra.communication.mail import send_mail
from orchestra.communication.slack import OrchestraSlackService
from orchestra.models import CommunicationPreference
from orchestra.models import StaffingRequestInquiry
from orchestra.models import Task
from orchestra.models import Project
from orchestra.utils.decorators import run_if
from orchestra.utils.task_properties import assignment_history
from orchestra.utils.task_properties import current_assignment


# TODO(jrbotros): design HTML template
def notify_status_change(
        task, previous_status=None, staffing_request_inquiry=None):
    """
    Notify workers after task has changed state
    """
    task_assignments = assignment_history(task)
    current_task_assignment = current_assignment(task)
    current_worker = None
    if current_task_assignment:
        current_worker = current_task_assignment.worker
    message_info = None

    # Notify worker when task initially picked up
    if task.status == Task.Status.PROCESSING:
        message_info = {
            'subject': "You've been assigned to a new task!",
            'message': ("You've been assigned to a new task. We can't wait "
                        "to see the great things you'll do!"),
            'recipient_list': [current_worker.user.email]
        }
    # Notify worker when assignment selected for review
    elif task.status == Task.Status.PENDING_REVIEW:
        message_info = {
            'subject': 'Your task is under review!',
            'message': ('Thanks for all your hard work, {}! The following '
                        'task was randomly selected for review by another '
                        'expert; you should hear back soon!').format(
                            current_worker.user.username),
            'recipient_list': [current_worker.user.email]
        }
    # Notify worker when assignment rejected
    elif task.status == Task.Status.POST_REVIEW_PROCESSING:
        message_info = {
            'subject': 'Your task has been returned',
            'message': ('Your reviewer sent back your task for a bit more '
                        'polish. Check out the feedback as soon as you can!'),
            'recipient_list': [current_worker.user.email]
        }
    # Notify all workers on a task when it has been completed
    elif task.status == Task.Status.COMPLETE:
        message_info = {
            'subject': 'Task complete!',
            'message': 'Congratulations! The task you worked on is complete.',
            'recipient_list': [assignment.worker.user.email
                               for assignment in task_assignments
                               if assignment.worker and
                               assignment.worker.user.email]
        }
    # Notify reviewer when task pending update is ready for re-review, but not
    # for a task moving from PENDING_REVIEW to REVIEWING
    elif (task.status == Task.Status.REVIEWING and
          previous_status == Task.Status.POST_REVIEW_PROCESSING):
        message_info = {
            'subject': 'A task is ready for re-review!',
            'message': ('A task has been updated and is ready for '
                        're-review!'),
            'recipient_list': [current_worker.user.email]
        }

    # Notify all workers on a task when it has been aborted
    elif task.status == Task.Status.ABORTED:
        message_info = {
            'subject': 'A task you were working on has been ended',
            'message': ('Unfortunately, the task you were working on has '
                        'been ended. Please reach out to us if you think this '
                        'has been done in error.'),
            'recipient_list': [assignment.worker.user.email
                               for assignment in task_assignments
                               if assignment.worker and
                               assignment.worker.user.email]
        }

    _notify_internal_slack_status_change(
        task, current_worker,
        staffing_request_inquiry=staffing_request_inquiry)
    if task.project.slack_group_id:
        _notify_experts_slack_status_change(
            task, current_worker,
            staffing_request_inquiry=staffing_request_inquiry)

    if message_info is not None:
        message_info['message'] += _task_information(task)
        comm_type = (CommunicationPreference.CommunicationType
                     .TASK_STATUS_CHANGE.value)
        send_mail(from_email=settings.ORCHESTRA_NOTIFICATIONS_FROM_EMAIL,
                  communication_type=comm_type,
                  fail_silently=True,
                  **message_info)


def notify_project_status_change(project):
    extra_explanation = ''
    if project.status == Project.Status.PAUSED:
        status_text = 'paused'
        extra_explanation = (
            'All activities will be put on hold until '
            'the project is reactivated.')
    elif project.status == Project.Status.ACTIVE:
        status_text = 'reactivated'
    elif project.status == Project.Status.COMPLETED:
        status_text = 'completed'
    elif project.status == Project.Status.ABORTED:
        status_text = 'aborted'
    else:
        return
    slack_message = (
        '*Project {} | {} has been {}.*\n'
        '{}'
        ).format(
            project.workflow_version.slug,
            project.short_description,
            status_text,
            extra_explanation)
    message_internal_slack_group(
        settings.SLACK_INTERNAL_NOTIFICATION_CHANNEL, slack_message)
    if project.slack_group_id:
        message_experts_slack_group(project.slack_group_id, slack_message)


def _task_information(task, with_slack_link=True):
    # TODO(jrbotros): incorporate Django sites framework
    dashboard_link = urljoin(settings.ORCHESTRA_URL,
                             reverse('orchestra:index'))
    task_information = ('\n\n'
                        'Project: {} | {}\n'
                        'Task: {} {}\n\n'
                        'View dashboard: {}\n').format(
                            task.project.workflow_version.slug,
                            task.project.short_description,
                            task.step.slug,
                            task.id,
                            dashboard_link)

    if task.project.slack_group_id and with_slack_link:
        slack_link = urljoin(settings.SLACK_EXPERTS_BASE_URL,
                             task.project.slack_group_id)
        task_information += 'View Slack channel: {}'.format(slack_link)

    return task_information


def _notify_slack_status_change(task, current_worker, slack_api_key,
                                slack_channel, with_slack_link=True,
                                with_user_mention=False,
                                staffing_request_inquiry=None):
    slack = OrchestraSlackService(slack_api_key)
    auto_staff_method = (
        StaffingRequestInquiry.CommunicationMethod.PREVIOUSLY_OPTED_IN)
    is_auto_staffed = (staffing_request_inquiry
                       and staffing_request_inquiry.communication_method
                       == auto_staff_method.value)
    processing_status_message = (
        'Task has been auto-staffed to a worker.'
        if is_auto_staffed else
        'Task has been picked up by a worker.')
    slack_statuses = {
        Task.Status.PROCESSING: processing_status_message,
        Task.Status.PENDING_REVIEW: 'Task is awaiting review.',
        Task.Status.REVIEWING: 'Task is under review.',
        Task.Status.POST_REVIEW_PROCESSING: 'Task was returned by reviewer.',
        Task.Status.COMPLETE: 'Task has been completed.',
        Task.Status.ABORTED: 'Task has been aborted.',
    }
    worker_string = current_worker.user.username if current_worker else None
    slack_user_id = current_worker.slack_user_id if current_worker else None
    if slack_user_id and with_user_mention:
        worker_string += ' (<@{}>)'.format(slack_user_id)
    slack_message = ('*{}*\n'
                     '>>>'
                     'Current worker: {}'
                     '{}').format(slack_statuses[task.status],
                                  worker_string,
                                  _task_information(
                                      task, with_slack_link=with_slack_link))
    slack.chat.post_message(slack_channel, slack_message)


@run_if('ORCHESTRA_SLACK_INTERNAL_ENABLED')
def _notify_internal_slack_status_change(
        task, current_worker, staffing_request_inquiry=None):
    _notify_slack_status_change(
        task, current_worker,
        settings.SLACK_INTERNAL_API_KEY,
        settings.SLACK_INTERNAL_NOTIFICATION_CHANNEL,
        staffing_request_inquiry=staffing_request_inquiry)


@run_if('ORCHESTRA_SLACK_EXPERTS_ENABLED')
def _notify_experts_slack_status_change(
        task, current_worker, staffing_request_inquiry=None):
    _notify_slack_status_change(
        task, current_worker,
        settings.SLACK_EXPERTS_API_KEY,
        task.project.slack_group_id,
        with_slack_link=False,
        with_user_mention=True,
        staffing_request_inquiry=staffing_request_inquiry)


@run_if('ORCHESTRA_SLACK_EXPERTS_ENABLED')
def message_experts_slack_group(slack_channel, slack_message):
    slack = OrchestraSlackService(settings.SLACK_EXPERTS_API_KEY)
    slack.chat.post_message(slack_channel, slack_message)


@run_if('ORCHESTRA_SLACK_INTERNAL_ENABLED')
def message_internal_slack_group(slack_channel, slack_message):
    slack = OrchestraSlackService(settings.SLACK_INTERNAL_API_KEY)
    slack.chat.post_message(slack_channel, slack_message)



from rest_framework import permissions

from orchestra.models import Step
from orchestra.models import Todo
from orchestra.models import Worker
from orchestra.project_api.auth import SignedUser
from orchestra.utils.notifications import message_experts_slack_group


class IsAssociatedWorker(permissions.BasePermission):
    """
    Permission for objects with `worker` field. Checks if request.user matches
    worker on object.
    """

    def has_object_permission(self, request, view, obj):
        worker = Worker.objects.get(user=request.user)
        return obj.worker == worker


def get_changed_fields(old_todo, new_todo):
    # To avoid Slack noise, only care about title and details
    changed_fields = []
    if old_todo.title != new_todo.title:
        changed_fields.append('title')
    if old_todo.details != new_todo.details:
        changed_fields.append('details')
    if len(changed_fields) > 1:
        changed_fields[-1] = 'and {}'.format(changed_fields[-1])
    return ' '.join(changed_fields)


def get_relevance_and_completion_changes(old_todo, new_todo):
    # When activity_log is updated, `todo_change = ''`
    # to avoid triggering any slack messages
    todo_change = ''
    if new_todo.status != old_todo.status:
        if new_todo.status == Todo.Status.COMPLETED.value and \
                old_todo.status == Todo.Status.PENDING.value:
            todo_change = 'complete'
        elif new_todo.status == Todo.Status.PENDING.value and \
                old_todo.status == Todo.Status.COMPLETED.value:
            todo_change = 'incomplete'
        elif new_todo.status == Todo.Status.DECLINED.value:
            todo_change = 'not relevant'
        elif old_todo.status == Todo.Status.DECLINED.value:
            todo_change = 'relevant'
    return todo_change


def get_update_message(old_todo, new_todo, sender=None):
    message = ''
    changes = get_relevance_and_completion_changes(old_todo, new_todo)
    changed_fields = get_changed_fields(old_todo, new_todo)
    head = '{} has updated `{}`:'.format(sender, new_todo.title) \
        if sender else '`{}` has been updated:'.format(new_todo.title)
    body = 'marked {}'.format(changes) if changes else ''
    tail = 'changed {}'.format(changed_fields) if changed_fields else ''
    if body or tail:
        tail = ', {}'.format(tail) if body and tail else '{}'.format(tail)
        message = '{} {}{}'.format(head, body, tail)
    return message


def _get_sender(user):
    if isinstance(user, SignedUser):
        return None
    return Worker.objects.get(
        user=user).formatted_slack_username()


def notify_single_todo_update(user, old_todo, todo):
    # To avoid Slack noise, only send updates for changed TODOs with
    # depth 0 (no parent) or 1 (no grandparent).
    sender = _get_sender(user)
    message = get_update_message(old_todo, todo, sender)
    if message and \
            (not (todo.parent_todo and todo.parent_todo.parent_todo)):
        message_experts_slack_group(
            todo.project.slack_group_id, message)


def notify_todo_created(todo, user):
    sender = _get_sender(user)
    tasks = (
        task for task in todo.project.tasks.all()
        if task.step.slug == todo.step.slug)
    recipients = ' & '.join(
        assignment.worker.formatted_slack_username()
        for task in tasks
        for assignment in task.assignments.all()
        if assignment and assignment.worker)
    if sender:
        message = '{} has created a new todo `{}` for {}.'.format(
            sender,
            todo.title,
            recipients if recipients else '`{}`'.format(todo.step.slug))
    else:
        message = 'A new todo `{}` was created for {}.'.format(
            todo.title,
            recipients if recipients else '`{}`'.format(todo.step.slug))
    message_experts_slack_group(
        todo.project.slack_group_id, message)


def get_step_by_project_id_and_step_slug(project_id, step_slug):
    step = Step.objects.get(
        slug=step_slug,
        workflow_version__projects__id=project_id)
    return step



# Generated by Django 2.2.13 on 2021-06-01 10:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0094_rename_team_messages_url'),
    ]

    operations = [
        migrations.AddField(
            model_name='todo',
            name='slug',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-27 14:12
from __future__ import unicode_literals

import django.db.models.deletion
import django.utils.timezone
from django.db import migrations
from django.db import models

import orchestra.models.communication.mixins
import orchestra.utils.models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0046_auto_20160426_2034'),
    ]

    operations = [
        migrations.CreateModel(
            name='StaffingRequest',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(
                    default=django.utils.timezone.now)),
                ('is_deleted', models.BooleanField(default=False)),
                ('request_cause', models.IntegerField(choices=[
                 (0, 'user'), (1, 'autostaff'), (2, 'restaff')])),
                ('project_description', models.TextField(blank=True, null=True)),
                ('task', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='orchestra.Task')),
                ('worker', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='orchestra.Worker')),
            ],
            options={
                'abstract': False,
            },
            bases=(orchestra.models.communication.mixins.StaffingRequestInquiryMixin,
                   orchestra.utils.models.DeleteMixin, models.Model),
        ),
        migrations.CreateModel(
            name='StaffingResponse',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(
                    default=django.utils.timezone.now)),
                ('is_deleted', models.BooleanField(default=False)),
                ('response_text', models.TextField(blank=True, null=True)),
                ('is_available', models.BooleanField()),
                ('is_winner', models.NullBooleanField()),
                ('request', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                              related_name='responses', to='orchestra.StaffingRequest')),
            ],
            options={
                'abstract': False,
            },
            bases=(orchestra.models.communication.mixins.StaffingResponseMixin,
                   orchestra.utils.models.DeleteMixin, models.Model),
        ),
        migrations.AddField(
            model_name='communicationpreference',
            name='created_at',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AddField(
            model_name='communicationpreference',
            name='is_deleted',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 2.2.13 on 2020-09-15 14:52

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0087_todo_task_data_migration'),
    ]

    operations = [
        migrations.AlterField(
            model_name='todo',
            name='task',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='todos_deprecated', to='orchestra.Task'),
        ),
    ]



import logging
from urllib.parse import urljoin

from django.conf import settings
from django import urls
from django.utils import timezone

from orchestra.communication.slack import OrchestraSlackService
from orchestra.models import Project
from orchestra.models import Worker
from orchestra.project_api.api import get_project_information

logger = logging.getLogger(__name__)


def project_management_information(project_id):
    project = Project.objects.get(id=project_id)
    project_information = get_project_information([project.id])
    project_information[project.id]['project']['status'] = dict(
        Project.STATUS_CHOICES).get(project.status, None)
    project_information[project.id]['project']['admin_url'] = urljoin(
        settings.ORCHESTRA_URL,
        urls.reverse(
            'admin:orchestra_project_change',
            args=(project_id,)))

    for slug, task in project_information[project.id]['tasks'].items():
        task['admin_url'] = urljoin(
            settings.ORCHESTRA_URL,
            urls.reverse(
                'admin:orchestra_task_change',
                args=(task['id'],)))

        for assignment in task['assignments']:
            assignment['admin_url'] = urljoin(
                settings.ORCHESTRA_URL,
                urls.reverse(
                    'admin:orchestra_taskassignment_change',
                    args=(assignment['id'],)))

            for iteration in assignment['iterations']:
                iteration['admin_url'] = urljoin(
                    settings.ORCHESTRA_URL,
                    urls.reverse(
                        'admin:orchestra_iteration_change',
                        args=(iteration['id'],)))
                if not iteration['end_datetime']:
                    iteration['end_datetime'] = timezone.now().isoformat()

    return project_information


def edit_slack_membership(project_id, username, action):
    slack = OrchestraSlackService()
    slack_user_id = Worker.objects.get(user__username=username).slack_user_id
    slack_group_id = Project.objects.get(id=project_id).slack_group_id
    if action == 'add':
        slack.conversations.invite(slack_group_id, slack_user_id)
    elif action == 'remove':
        slack.conversations.kick(slack_group_id, slack_user_id)
    else:
        raise Exception('Action not found.')



from django.conf.urls import url
from django.views.generic import RedirectView

from orchestra.accounts.views import AccountSettingsView
from orchestra.accounts.views import AvailabilitySettingsView
from orchestra.accounts.views import CommunicationPreferenceSettingsView
from orchestra.accounts.views import OrchestraRegistrationView

urlpatterns = [
    # We have to override the register endpoint to emit the
    # `orchestra_user_registered` signal
    url(r'^accounts/register/$',
        OrchestraRegistrationView.as_view(),
        name='registration_register'),
    url(r'^accounts/settings/$',
        AccountSettingsView.as_view(), name='account_settings'),
    url(r'^accounts/communication_preference_settings/$',
        CommunicationPreferenceSettingsView.as_view(),
        name='communication_preference_settings'),
    url(r'^accounts/availability_settings/$',
        AvailabilitySettingsView.as_view(),
        name='availability_settings'),
    url(r'^password/change/done/$',
        RedirectView.as_view(
            pattern_name='account_settings'),
        name='auth_password_change_done'),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.10 on 2018-06-15 16:06
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import jsonfield.fields
import orchestra.models.core.mixins
import orchestra.utils.models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0075_add_sanity_checks'),
    ]

    operations = [
        migrations.CreateModel(
            name='TodoListTemplate',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(
                    default=django.utils.timezone.now)),
                ('is_deleted', models.BooleanField(default=False)),
                ('slug', models.CharField(max_length=200, unique=True)),
                ('name', models.CharField(max_length=200)),
                ('description', models.TextField()),
                ('todos', jsonfield.fields.JSONField(default={'items': []})),
                ('creator', models.ForeignKey(blank=True, null=True,
                                              on_delete=django.db.models.deletion.SET_NULL, related_name='creator', to='orchestra.Worker')),
            ],
            bases=(orchestra.models.core.mixins.TodoListTemplateMixin,
                   orchestra.utils.models.DeleteMixin, models.Model),
        ),
        migrations.AddField(
            model_name='todo',
            name='activity_log',
            field=jsonfield.fields.JSONField(default={'actions': []}),
        ),
        migrations.AddField(
            model_name='todo',
            name='parent_todo',
            field=models.ForeignKey(
                null=True, on_delete=django.db.models.deletion.CASCADE, related_name='parent', to='orchestra.Todo'),
        ),
        migrations.AddField(
            model_name='todo',
            name='skipped_datetime',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='todo',
            name='template',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    related_name='template', to='orchestra.TodoListTemplate'),
        ),
    ]



# Generated by Django 2.2.13 on 2021-02-18 18:39

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import jsonfield.fields
import orchestra.models.core.mixins
import orchestra.utils.datetime_utils
import orchestra.utils.models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0091_auto_20210121_1952'),
    ]

    operations = [
        migrations.AddField(
            model_name='step',
            name='assignable_hours_function',
            field=jsonfield.fields.JSONField(default={}),
        ),
        migrations.AddField(
            model_name='worker',
            name='max_autostaff_hours_per_day',
            field=models.FloatField(blank=True, default=0),
        ),
        migrations.AlterField(
            model_name='staffbotrequest',
            name='request_cause',
            field=models.IntegerField(choices=[(0, 'user'), (1, 'task policy'), (2, 'restaff')]),
        ),
        migrations.AlterField(
            model_name='staffingrequestinquiry',
            name='communication_method',
            field=models.IntegerField(choices=[(0, 'slack'), (1, 'email'), (2, 'already opted in')]),
        ),
        migrations.CreateModel(
            name='WorkerAvailability',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now)),
                ('is_deleted', models.BooleanField(default=False)),
                ('week', models.DateField(default=orchestra.utils.datetime_utils.first_day_of_the_week)),
                ('hours_available_mon', models.FloatField(default=0, verbose_name='Monday')),
                ('hours_available_tues', models.FloatField(default=0, verbose_name='Tuesday')),
                ('hours_available_wed', models.FloatField(default=0, verbose_name='Wednesday')),
                ('hours_available_thurs', models.FloatField(default=0, verbose_name='Thursday')),
                ('hours_available_fri', models.FloatField(default=0, verbose_name='Friday')),
                ('hours_available_sat', models.FloatField(default=0, verbose_name='Saturday')),
                ('hours_available_sun', models.FloatField(default=0, verbose_name='Sunday')),
                ('worker', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='availabilities', to='orchestra.Worker')),
            ],
            bases=(orchestra.models.core.mixins.WorkerAvailabilityMixin, orchestra.utils.models.DeleteMixin, models.Model),
        ),
        migrations.AddConstraint(
            model_name='workeravailability',
            constraint=models.UniqueConstraint(fields=('worker', 'week'), name='uniq_worker_week'),
        ),
    ]



from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from django.db import IntegrityError

from orchestra.models import Todo
from orchestra.models import TodoQA
from orchestra.models import TodoListTemplate
from orchestra.json_schemas.todos import TodoListSchema
from orchestra.json_schemas.todos import TodoActionListSchema
from orchestra.utils.mixins import JSONSchemaValidationMixin
from orchestra.utils.common_helpers import get_step_by_project_id_and_step_slug


class TodoQASerializer(serializers.ModelSerializer):

    class Meta:
        model = TodoQA
        fields = (
            'id',
            'created_at',
            'todo',
            'approved',
            'comment')
        read_only_fields = ('id',)


class TodoListTemplateSerializer(serializers.ModelSerializer,
                                 JSONSchemaValidationMixin):
    json_schemas = {
        'todos': TodoListSchema
    }

    class Meta:
        model = TodoListTemplate
        fields = (
            'id',
            'created_at',
            'slug',
            'name',
            'description',
            'creator',
            'todos')
        read_only_fields = ('id',)


class TodoBulkCreateListSerializer(serializers.ListSerializer):
    def create(self, validated_data):
        result = [self.child.create(attrs) for attrs in validated_data]
        try:
            self.child.Meta.model.objects.bulk_create(result)
        except IntegrityError as e:
            raise ValidationError(e)
        return result

    def update(self, instances, validated_data):
        result = []
        for instance, data in zip(instances, validated_data):
            result.append(self.child.update(instance, data))

        writable_fields = [
            x for x in self.child.Meta.fields
            if x not in self.child.Meta.read_only_fields
        ]

        try:
            self.child.Meta.model.objects.bulk_update(result, writable_fields)
        except IntegrityError as e:
            raise ValidationError(e)

        return result


class StepField(serializers.Field):
    def to_representation(self, data):
        return data.slug

    def to_internal_value(self, value):
        return value


class BulkTodoSerializer(serializers.ModelSerializer):
    json_schemas = {
        'activity_log': TodoActionListSchema
    }

    step = StepField()
    additional_data = serializers.JSONField(required=False)

    # TODO(murat): Remove this validation when project
    # becomes a required field in models
    def validate(self, data):
        step = data.get('step')
        project = data.get('project')
        if step is not None and project is None:
            raise serializers.ValidationError({
                'project': [
                    'if step is given, project should also be supplied.']})
        return data

    def _set_step_to_validated_data(self, validated_data):
        project_id = validated_data['project'].id
        step = get_step_by_project_id_and_step_slug(project_id,
                                                    validated_data['step'])
        validated_data['step'] = step
        return validated_data

    def create(self, validated_data):
        validated_data = self._set_step_to_validated_data(validated_data)
        instance = Todo(**validated_data)
        if isinstance(self._kwargs['data'], dict):
            instance.save()
        return instance

    def update(self, instance, validated_data):
        if validated_data.get('step') is not None:
            validated_data = self._set_step_to_validated_data(validated_data)
        for k, v in validated_data.items():
            setattr(instance, k, v)

        if isinstance(self._kwargs['data'], dict):
            instance.save()
        return instance

    class Meta:
        model = Todo
        fields = (
            'id',
            'created_at',
            'title',
            'details',
            'parent_todo',
            'template',
            'start_by_datetime',
            'due_datetime',
            'activity_log',
            'section',
            'project',
            'step',
            'slug',
            'order',
            'status',
            'additional_data',
            'required',
            'is_deleted')
        read_only_fields = ('id',)
        list_serializer_class = TodoBulkCreateListSerializer


class BulkTodoSerializerWithoutQA(BulkTodoSerializer):
    qa = serializers.SerializerMethodField()

    def get_qa(self, obj):
        return None

    class Meta(BulkTodoSerializer.Meta):
        fields = BulkTodoSerializer.Meta.fields + ('qa',)
        read_only_fields = ('id',)


class BulkTodoSerializerWithQA(BulkTodoSerializer):
    qa = TodoQASerializer(read_only=True)

    class Meta(BulkTodoSerializer.Meta):
        fields = BulkTodoSerializer.Meta.fields + ('qa',)
        read_only_fields = ('id',)



import logging
import copy

from django.contrib import messages
from django.contrib.admin.views.decorators import staff_member_required
from django.shortcuts import redirect
from django.shortcuts import render
from django.utils.decorators import method_decorator
from django.views.generic import View
from rest_framework import generics
from rest_framework import permissions
from rest_framework.viewsets import ModelViewSet
from rest_framework.response import Response
from rest_framework.decorators import action
from jsonview.exceptions import BadRequest
from django.db.models.query import QuerySet

from orchestra.core.errors import TodoListTemplateValidationError
from orchestra.models import Task
from orchestra.models import Todo
from orchestra.models import TodoQA
from orchestra.models import TodoListTemplate
from orchestra.todos.forms import ImportTodoListTemplateFromSpreadsheetForm
from orchestra.todos.filters import QueryParamsFilterBackend
from orchestra.todos.serializers import BulkTodoSerializer
from orchestra.todos.serializers import BulkTodoSerializerWithoutQA
from orchestra.todos.serializers import BulkTodoSerializerWithQA
from orchestra.todos.serializers import TodoQASerializer
from orchestra.todos.serializers import TodoListTemplateSerializer
from orchestra.utils.common_helpers import notify_todo_created
from orchestra.utils.common_helpers import notify_single_todo_update
from orchestra.todos.api import add_todolist_template
from orchestra.utils.decorators import api_endpoint
from orchestra.todos.auth import IsAssociatedWithTodosProject
from orchestra.todos.auth import IsAssociatedWithProject
from orchestra.todos.auth import IsAssociatedWithTask
from orchestra.todos.import_export import import_from_spreadsheet

logger = logging.getLogger(__name__)


@api_endpoint(methods=['POST'],
              permissions=(IsAssociatedWithProject,),
              logger=logger)
def update_todos_from_todolist_template(request):
    todolist_template_slug = request.data.get('todolist_template')
    project_id = request.data.get('project')
    step_slug = request.data.get('step')
    try:
        add_todolist_template(todolist_template_slug, project_id, step_slug)
        todos = Todo.objects.filter(
            project__id=project_id).order_by('-created_at')
        serializer = BulkTodoSerializerWithoutQA(todos, many=True)
        return Response(serializer.data)
    except TodoListTemplate.DoesNotExist:
        raise BadRequest('TodoList Template not found for the given slug.')


@api_endpoint(methods=['GET'],
              permissions=(permissions.IsAuthenticated,
                           IsAssociatedWithTask),
              logger=logger)
def worker_task_recent_todo_qas(request):
    """
    Returns TodoQA recommendations for the requesting user and task.

    The TodoQAs for the recommendation are selected using the following logic:
    1. If the given task has TodoQAs, use them.
    2. Otherwise, use the TodoQAs from the requesting user's most recent
    task with matching task slug.
    """

    task_id = request.query_params.get('task')
    task_todo_qas = TodoQA.objects.filter(todo__project__tasks__id=task_id)

    if task_todo_qas.exists():
        todo_qas = task_todo_qas
    else:
        task = Task.objects.get(pk=task_id)
        most_recent_worker_task_todo_qa = TodoQA.objects.filter(
            todo__project__tasks__assignments__worker__user=request.user,
            todo__step__slug=task.step.slug
        ).order_by('-created_at').first()

        if most_recent_worker_task_todo_qa:
            project = most_recent_worker_task_todo_qa.todo.project
            step = most_recent_worker_task_todo_qa.todo.step
            todo_qas = TodoQA.objects.filter(
                todo__project=project,
                todo__step=step,
                approved=False)
        else:
            todo_qas = TodoQA.objects.none()

    todos_recommendation = {todo_qa.todo.title: TodoQASerializer(
        todo_qa).data for todo_qa in todo_qas}
    return Response(todos_recommendation)


class TodoQADetail(generics.UpdateAPIView):
    permission_classes = (permissions.IsAuthenticated,
                          IsAssociatedWithTodosProject)

    serializer_class = TodoQASerializer
    queryset = TodoQA.objects.all()


class TodoQAList(generics.CreateAPIView):
    permission_classes = (permissions.IsAuthenticated,
                          IsAssociatedWithProject)

    serializer_class = TodoQASerializer
    queryset = TodoQA.objects.all()


class TodoListTemplateDetail(generics.RetrieveUpdateAPIView):
    permission_classes = (permissions.IsAuthenticated,)

    serializer_class = TodoListTemplateSerializer
    queryset = TodoListTemplate.objects.all()


class TodoListTemplateList(generics.ListCreateAPIView):
    permission_classes = (permissions.IsAuthenticated,)

    serializer_class = TodoListTemplateSerializer
    queryset = TodoListTemplate.objects.all()


class GenericTodoViewset(ModelViewSet):
    """
    A base viewset inherited by multiple viewsets, created
    to de-duplicate Todo-related views.
    It lacks permission and auth classes, so that child
    classes can select their own.
    """
    serializer_class = BulkTodoSerializer
    filter_backends = (QueryParamsFilterBackend,)
    # Note: additional_data__nested_field is not supported in filterset_fields
    # This issue can be fixed when we migrate to Django 3.1
    # and convert additional_data from django-jsonfields to the native one.
    filterset_fields = ('project__id', 'step__slug', 'id__in')
    queryset = Todo.objects.select_related('step', 'qa').all()

    def get_serializer(self, *args, **kwargs):
        if isinstance(kwargs.get('data', {}), list):
            kwargs['many'] = True

        return super().get_serializer(*args, **kwargs)

    def get_queryset(self, ids=None):
        queryset = super().get_queryset()
        if ids is not None:
            queryset = queryset.filter(id__in=ids)
        return queryset.order_by('-created_at')

    @action(detail=False, methods=['delete'])
    def delete(self, request, *args, **kwargs):
        data = self.get_queryset(ids=request.data).delete()
        return Response(data)

    @action(detail=False, methods=['put'])
    def put(self, request, *args, **kwargs):
        partial = kwargs.get('partial', False)
        ids = [x['id'] for x in request.data]
        # Sort the queryset and data by primary key
        # so we update the correct records.
        sorted_data = sorted(request.data, key=lambda x: x['id'])
        instances = self.get_queryset(ids=ids).order_by('id')
        serializer = self.get_serializer(
            instances, data=sorted_data, partial=partial, many=True)
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)
        data = serializer.data
        return Response(data)

    @action(detail=False, methods=['patch'])
    def patch(self, request, *args, **kwargs):
        kwargs['partial'] = True
        return self.put(request, *args, **kwargs)

    def perform_update(self, serializer):
        old_data = copy.deepcopy(serializer.instance)
        old_todos = list(old_data) if isinstance(
            old_data, QuerySet) else [old_data]
        data = serializer.save()
        todos = data if isinstance(data, list) else [data]
        for old_todo, new_todo in zip(old_todos, todos):
            if isinstance(new_todo, Todo):
                notify_single_todo_update(
                    self.request.user, old_todo, new_todo)

    def perform_create(self, serializer):
        data = serializer.save()
        todos = data if isinstance(data, list) else [data]
        for todo in todos:
            if isinstance(todo, Todo):
                notify_todo_created(todo, self.request.user)


class TodoViewset(GenericTodoViewset):
    """
    This viewset inherits from GenericTodoViewset is used by two endpoints
    (see urls.py).
    todo/ -- For creating and listing Todos.
    todo/1234/ -- For updating a Todo.
    """
    http_method_names = ['get', 'post', 'put', 'delete']

    def get_permissions(self):
        permission_classes = (permissions.IsAuthenticated,
                              IsAssociatedWithProject)
        if self.action == 'update':
            permission_classes = (permissions.IsAuthenticated,
                                  IsAssociatedWithTodosProject)
        return [permission() for permission in permission_classes]

    def get_serializer_class(self):
        if self.action == 'list' or self.action == 'create':
            # Only include todo QA data for users in the
            # `project_admins` group.
            if self.request.user.groups.filter(
                    name='project_admins').exists():
                return BulkTodoSerializerWithQA
            else:
                return BulkTodoSerializerWithoutQA
        else:
            return super().get_serializer_class()


@method_decorator(staff_member_required, name='dispatch')
class ImportTodoListTemplateFromSpreadsheet(View):
    TEMPLATE = 'orchestra/import_todo_list_template_from_spreadsheet.html'

    def get(self, request, pk):
        # Display a form with a spreadsheet URL for import.
        return render(
            request,
            self.TEMPLATE,
            {'form': ImportTodoListTemplateFromSpreadsheetForm(initial={})})

    def post(self, request, pk):
        # Try to import the spreadsheet and redirect back to the admin
        # entry for the imported TodoListTemplate. If we encounter an
        # error, display the error on the form.
        form = ImportTodoListTemplateFromSpreadsheetForm(request.POST)
        context = {'form': form}
        if form.is_valid():
            try:
                todo_list_template = TodoListTemplate.objects.get(
                    id=pk)
                import_from_spreadsheet(
                    todo_list_template,
                    form.cleaned_data['spreadsheet_url'],
                    request)
                messages.info(
                    request, 'Successfully imported from spreadsheet.')
                return redirect(
                    'admin:orchestra_todolisttemplate_change',
                    pk)
            except TodoListTemplateValidationError as e:
                context['import_error'] = str(e)
        else:
            context['import_error'] = 'Please provide a spreadsheet URL'
        return render(
            request,
            self.TEMPLATE,
            context)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-14 21:18
from __future__ import unicode_literals

import django.utils.timezone
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0036_remove_taskassignment_snapshots'),
    ]

    operations = [
        migrations.AddField(
            model_name='iteration',
            name='created_at',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AddField(
            model_name='iteration',
            name='is_deleted',
            field=models.BooleanField(default=False),
        ),
    ]



from orchestra.models.communication.models import CommunicationPreference
from orchestra.models.communication.models import StaffBotRequest
from orchestra.models.communication.models import StaffingRequestInquiry
from orchestra.models.communication.models import StaffingResponse
from orchestra.models.core.models import Workflow
from orchestra.models.core.models import WorkflowVersion
from orchestra.models.core.models import Certification
from orchestra.models.core.models import SanityCheck
from orchestra.models.core.models import Step
from orchestra.models.core.models import Worker
from orchestra.models.core.models import WorkerCertification
from orchestra.models.core.models import WorkerAvailability
from orchestra.models.core.models import Project
from orchestra.models.core.models import Task
from orchestra.models.core.models import TaskAssignment
from orchestra.models.core.models import Iteration
from orchestra.models.core.models import TimeEntry
from orchestra.models.core.models import TaskTimer
from orchestra.models.core.models import Todo
from orchestra.models.core.models import TodoQA
from orchestra.models.core.models import TodoListTemplate
from orchestra.models.core.models import TodoListTemplateImportRecord
from orchestra.models.core.models import PayRate


__all__ = [
    'CommunicationPreference',
    'StaffBotRequest',
    'StaffingRequestInquiry',
    'StaffingResponse',
    'Workflow',
    'WorkflowVersion',
    'Certification',
    'SanityCheck',
    'Step',
    'Worker',
    'WorkerCertification',
    'WorkerAvailability',
    'Project',
    'Task',
    'TaskAssignment',
    'Iteration',
    'TimeEntry',
    'TaskTimer',
    'Todo',
    'TodoQA',
    'TodoListTemplate',
    'TodoListTemplateImportRecord',
    'PayRate',
]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.12 on 2018-10-02 18:47
from __future__ import unicode_literals

from django.db import migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0081_step_todolist_templates_to_apply'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='tags',
            field=jsonfield.fields.JSONField(default={'tags': []}),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.12 on 2018-07-10 21:18
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import orchestra.models.core.mixins
import orchestra.utils.models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0077_auto_20180629_1827'),
    ]

    operations = [
        migrations.CreateModel(
            name='TodoQA',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now)),
                ('is_deleted', models.BooleanField(default=False)),
                ('comment', models.TextField(blank=True, null=True)),
                ('approved', models.NullBooleanField()),
                ('todo', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='qa', to='orchestra.Todo')),
            ],
            bases=(orchestra.models.core.mixins.TodoQAMixin, orchestra.utils.models.DeleteMixin, models.Model),
        ),
    ]
